/*R: current memory requirements for GPU part:
gpu_chirps ->32k*4*2*16=4M
ocl_global_buf* ->32k*4*2*32=8M

If unroll by data chunk will be used, then GPU mem requirements will be:
gpu_chirps + N_of_unrolls*(ocl_global_buf1+ocl_global_buf2)=4M+N*16M
for mem size of 128M 7 unrolls are possible.
*/

//#pragma OPENCL EXTENSION cl_amd_printf : enable

#define COMBINED_DECHIRP_KERNEL 1 //R:To use (1) or not to use (0) Frizz's compute-intencive dechirping kernel
#define NO_BESTS 1 //Will govern best pulses search inside find single pulse kernel

/*R: 
This defines enable petri33's optimizations of split and dechirp kernels.
Don't forget to delete kernels binary cache when switching define.
This optimization behaves differently on different GPUs so disabled by default for ATi and Intel GPUs now.
It known to be faster on some NV GPUs, gives good speedup for AMD Trinity but
gives slowdown for some AMD discrete GPUs (GCN). 
Profiled on Loveland: almost twice slowdown on split kernel, no change on dechirp kernel.
Profiled on pre-FERMI GSO9600: diminishing speedup on split, same speed on dechirp
Profiled on pre-FERMI GT9400: speedup on both kernels
*/

#if USE_OPENCL_NV
	#define PCHIRP 1
	#define PSPLIT 1
#else
	#define PCHIRP 0
	#define PSPLIT 0
#endif

/*R: 
This define enables (if set to 1) or disables (set to 0) petri33's modification of SinglePulse kernels.
Don't forget to delete binary cache when switching it.
Initially used for Kepler-class NV GPUs. Needs additional testing on other hardware.
Profiled on Loveland GPU - no improvement for this type of devices. 
Profiled on pre-FERMI GT9400 and GSO9600: both show good speedup on first SingleFind kernel and slowdown on second one.
*/
#if USE_OPENCL_NV
	#define PSFIND0 1
	#define PSFIND1 0
#else
	#define PSFIND0 0
	#define PSFIND1 0
#endif

__kernel __attribute__((work_group_size_hint(1, 1, 1)))
void set_mem_kernel_cl(__global float* data){
	uint tid=get_global_id(0);
	data[tid]=0.0f;
}
__kernel 
void set_guard_elements(__global float* data, const int offset){
	uint tid=get_global_id(0);
	data[offset+tid]=0.0f;
}
#if PSPLIT
__kernel void splitter_bits_to_float_range_kernel(__global uint4* gpu_raw,__global float4* gpu_data,const uint raw_offset){
//R: I assume that first bit represents data[0].re, second - data[0].im and so on.
//R: Each workitem takes 4 unsigned ints from bit stream, that is 4*4*8 bits or 4*4*8/2 data elements
//R: that is, 64 data elements per work item. We need 32k*DATA_CHUNK_UNROLL elements total =>
//R: there are plenty workitems to load (512*7 in current version) all processing elements of GPU.
//R: raw_offset contains offset into raw array considering raw as array of uint4 entities
//R: (that is, each element contains 4*4*8 bits)
	uint tid= get_global_id(0);
	uint dchunk=get_global_id(1);
//R: we need 512 work items per data array
	int raw_offset_fin=raw_offset+tid+512/2*dchunk;
	if(raw_offset_fin>=(8*1024*1024>>(2+2))) return;
	uint4 bits=gpu_raw[raw_offset_fin];//R:data chunks overlapped by 50% hence 512/2
	int data_offset = dchunk*(32768/2)+tid*32;
	float4 f1, f2, f3, f4;

	f1.x = (bits.x & 1 ? 1 : -1); bits.x >>= 1;
	f1.y = (bits.x & 1 ? 1 : -1); bits.x >>= 1;
	f1.z = (bits.x & 1 ? 1 : -1); bits.x >>= 1;
	f1.w = (bits.x & 1 ? 1 : -1); bits.x >>= 1;

	f2.x = (bits.y & 1 ? 1 : -1); bits.y >>= 1;
	f2.y = (bits.y & 1 ? 1 : -1); bits.y >>= 1;
	f2.z = (bits.y & 1 ? 1 : -1); bits.y >>= 1;
	f2.w = (bits.y & 1 ? 1 : -1); bits.y >>= 1;

        for(uint i = 0; i < 8; i++)
          {
            bool bx, by, bz, bw;

            bx = (bool)(bits.z & 1), by = (bool)(bits.z & 2), bz = (bool)(bits.z & 4), bw = (bool)(bits.z & 8);
	    //bool bx = i1, by = i2, bz = i3, bw = i4;

            bits.z >>= 4;

	    f3.x = (bx ? 1 : -1); 
	    f3.y = (by ? 1 : -1); 
	    f3.z = (bz ? 1 : -1); 
	    f3.w = (bw ? 1 : -1); 

	    gpu_data[data_offset   ] = f1;

	    //i1 = bits.w & 1, i2 = bits.w & 2, i3 = bits.w & 4, i4 = bits.w & 8; 
	    //bx = i1, by = i2, bz = i3, bw = i4;
	    bx = (bool)(bits.w & 1); by = (bool)(bits.w & 2); bz = (bool)(bits.w & 4); bw = (bool)(bits.w & 8);

            bits.w >>= 4;

	    f4.x = (bx ? 1 : -1); 
	    f4.y = (by ? 1 : -1); 
	    f4.z = (bz ? 1 : -1); 
	    f4.w = (bw ? 1 : -1); 

	    gpu_data[data_offset +8] = f2;

	    //= bits.x & 1, i2 = bits.x & 2, i3 = bits.x & 4, i4 = bits.x & 8; 
	    //bx = i1, by = i2, bz = i3, bw = i4;
	    bx = (bool)(bits.x & 1); by = (bool)(bits.x & 2); bz = (bool)(bits.x & 4); bw = (bool)(bits.x & 8);

            bits.x >>= 4;

	    f1.x = (bx ? 1 : -1); 
	    f1.y = (by ? 1 : -1); 
	    f1.z = (bz ? 1 : -1); 
	    f1.w = (bw ? 1 : -1); 

	    gpu_data[data_offset+16] = f3;

//	    //= bits.y & 1, i2 = bits.y & 2, i3 = bits.y & 4, i4 = bits.y & 8; 
//	    //bx = i1, by = i2, bz = i3, bw = i4;
	    bx = (bool)(bits.y & 1); by = (bool)(bits.y & 2); bz = (bool)(bits.y & 4); bw = (bool)(bits.y & 8);

            bits.y >>= 4;

	    f2.x = (bx ? 1 : -1); 
	    f2.y = (by ? 1 : -1); 
	    f2.z = (bz ? 1 : -1); 
	    f2.w = (bw ? 1 : -1); 

	    gpu_data[data_offset+24] = f4;

	    data_offset++;
          }

}

#else
__kernel void splitter_bits_to_float_range_kernel(__global uint4* gpu_raw,__global float4* gpu_data,const uint raw_offset){
//R: I assume that first bit represents data[0].re, second - data[0].im and so on.
//R: Each workitem takes 4 unsigned ints from bit stream, that is 4*4*8 bits or 4*4*8/2 data elements
//R: that is, 64 data elements per work item. We need 32k*DATA_CHUNK_UNROLL elements total =>
//R: there are plenty workitems to load (512*7 in current version) all processing elements of GPU.
//R: raw_offset contains offset into raw array considering raw as array of uint4 entities
//R: (that is, each element contains 4*4*8 bits)
	uint tid= get_global_id(0);
	uint dchunk=get_global_id(1);
	float4 data;
//R: we need 512 work items per data array
	int raw_offset_fin=raw_offset+tid+512/2*dchunk;
	if(raw_offset_fin>=(8*1024*1024>>(2+2))) return;
	uint4 bits=gpu_raw[raw_offset_fin];//R:data chunks overlapped by 50% hence 512/2
	uint data_offset=dchunk*(32768/2)+tid*32;
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		data.y=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		data.z=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		data.w=2.0f*(bits.x & 1)-1.0f;bits.x>>=1;
		gpu_data[data_offset]=data;
	}
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		data.y=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		data.z=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		data.w=2.0f*(bits.y & 1)-1.0f;bits.y>>=1;
		gpu_data[data_offset]=data;
	}
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		data.y=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		data.z=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		data.w=2.0f*(bits.z & 1)-1.0f;bits.z>>=1;
		gpu_data[data_offset]=data;
	}
	for(uint i=0;i<8;i++,data_offset++){
		data.x=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		data.y=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		data.z=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		data.w=2.0f*(bits.w & 1)-1.0f;bits.w>>=1;
		gpu_data[data_offset]=data;
	}
}
#endif


 // start of combined dechirp kernel if/else-swith
#if COMBINED_DECHIRP_KERNEL	// must be enabled by hand - not currently set on OpenCL compiler

// dechirping done entirely on GPU

#define sigma 1310.7200f
#define rcp_2SigmaSqr 0.000000291038304567337040f
#define normalize_val 1.05424785025598620f

#ifndef M_PI
	#define M_PI 3.14159265358979323846264338f
#endif

#define FFT_SIZE 32768

#if PCHIRP

inline float4 calc_chirp(float4 cconst, float dm_start, int i) 
{
  float4 result;
  float phase1, phase2;
	
  float phase_const = M_PI * (dm_start + (float)i);

  float freq1 = cconst.x * cconst.x;
  cconst.x *= 0.00176f;

  float freq2 = cconst.y * cconst.y;
  cconst.y *= 0.00176f;

  cconst.x += 1.0f;
  cconst.y += 1.0f;

  cconst.x = 1.0f/cconst.x;
  cconst.y = 1.0f/cconst.y;

  freq1 = freq1 * cconst.x;
  freq2 = freq2 * cconst.y;

  phase1 = phase_const * freq1;
  phase2 = phase_const * freq2;

  result.x = native_cos(phase1);
  result.y = native_sin(phase1);
  result.z = native_cos(phase2);	
  result.w = native_sin(phase2);	

  result.x *= cconst.z;
  result.y *= cconst.z;
  result.z *= cconst.w;
  result.w *= cconst.w;

  return result;
}

#else
inline float4 calc_chirp(int j, int i, float dm_start) 
{
	float4 result;
	
	float phase_const = M_PI*(dm_start + (float)i);
	
	float g;
	float freq;
	float multiplier;	
	float phase;

	if(	j < (FFT_SIZE/2) ) {	
	
		float float_j = (float)j;
		g = float_j + 0.5f;
	
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;
	
		freq = float_j/32768.0f;
	
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));
#if USE_OPENCL_INTEL
		float cosres;
		result.y=sincos(phase,&cosres)*multiplier;
		result.x=cosres*multiplier;
#else
		result.x = native_cos(phase) * multiplier;
		result.y = native_sin(phase) * multiplier;
#endif
		float_j = float_j +1.0f;
		g = float_j + 0.5f;
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;
		
		freq = float_j/32768.0f;
		
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));
#if USE_OPENCL_INTEL
		result.w=sincos(phase,&cosres)*multiplier;
		result.z=cosres*multiplier;
#else
		result.z = native_cos(phase) * multiplier;	
		result.w = native_sin(phase) * multiplier;	
#endif		
	} else 
	{
	
		j = j -16384;

		g = (float)(16383 -j) + 0.5f;
	
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;
	
		freq = (float)(j + 16384)/32768.0f -1.0f;
	
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));
#if USE_OPENCL_INTEL
		float cosres;
		result.y=sincos(phase,&cosres)*multiplier;
		result.x=cosres*multiplier;
#else
		result.x = native_cos(phase) * multiplier;
		result.y = native_sin(phase) * multiplier;
#endif
		j = j +1;
		g = (float)(16383 -j) + 0.5f;
		multiplier = native_sqrt(1.0f - native_exp(-(g * g * rcp_2SigmaSqr))) *normalize_val;	
		
		freq = (float)(j + 16384)/32768.0f -1.0f;
		
		phase = phase_const * freq * freq * (1.0f / (0.00176f *freq +1.0f));
#if USE_OPENCL_INTEL
		result.w=sincos(phase,&cosres)*multiplier;
		result.z=cosres*multiplier;
#else
		result.z = native_cos(phase) * multiplier;	
		result.w = native_sin(phase) * multiplier;	
#endif		
	}
	return result;
}
#endif
#if 0
__kernel void build_chirp_table_small_kernel_cl(__global float4* gpu_chirps,const float dm_start){
	int tid = get_global_id(0);
	int i=get_global_id(1);
	gpu_chirps[i*(FFT_SIZE/2)+tid]=calc_chirp(tid *2, i, dm_start);
}
__kernel void dechirp_range2_kernel ( __global float4* gpu_data ,
										__global float4* gpu_dechirped,
										__global float4* gpu_chirps)
{

	uint tid = get_global_id(0);
	uint dchunk=get_global_id(1);
	//R: each work item will process 2 complex data elements and write 2*2*16 new complex elements into dechirped array
	
	float4 data=gpu_data[tid+dchunk*(FFT_SIZE/2)];
	float4 cur_chirp;
	float4 cur_dechirp;
	float xx, yy, zz, ww, yx, xy, wz, zw;
	
	for(uint i=0;i<16;i++){//R: can be optimized via mad instruction probably
		
		cur_chirp =gpu_chirps[i*(FFT_SIZE/2)+tid];//calc_chirp(tid *2, i, dm_start);
		
		//negative sign
		xx = data.x*cur_chirp.x;
		yy = data.y*cur_chirp.y;
		yx = data.y*cur_chirp.x;
		xy = data.x*cur_chirp.y;
		zz = data.z*cur_chirp.z;
		ww = data.w*cur_chirp.w;
		wz = data.w*cur_chirp.z;
		zw = data.z*cur_chirp.w;
#if APV505
		cur_dechirp.x = xx - yy;
		cur_dechirp.y = yx + xy;
		cur_dechirp.z = zz - ww;
		cur_dechirp.w = wz + zw;
#else
		cur_dechirp.x = xx + yy;
		cur_dechirp.y = yx - xy;
		cur_dechirp.z = zz + ww;
		cur_dechirp.w = wz - zw;
#endif
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+0)*(FFT_SIZE/2)+tid]=cur_dechirp;

		//positive sign
#if APV505
		cur_dechirp.x = xx + yy;
		cur_dechirp.y = yx - xy;
		cur_dechirp.z = zz + ww;
		cur_dechirp.w = wz - zw;
#else
		cur_dechirp.x = xx - yy;
		cur_dechirp.y = yx + xy;
		cur_dechirp.z = zz - ww;
		cur_dechirp.w = wz + zw;
#endif
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+1)*(FFT_SIZE/2)+tid]=cur_dechirp;

		}
}
#else

#if PCHIRP
__kernel void dechirp_range1_kernel ( __global float4* gpu_data ,
										__global float4* gpu_dechirped,
										__const float dm_start)
{
  uint tid = get_global_id(0);
  uint dchunk=get_global_id(1);
  int j = tid*2;
  float4 cconst;
  float nrcp = -rcp_2SigmaSqr;
	
  float g1, g2;
  float freq1, freq2;
  float multiplier1, multiplier2;	

  if(j < (FFT_SIZE/2)) 
    {	
      g1 = (float)j + 0.5f;
      g2 = (float)j + 1.5f;
      float float_j1 = (float)j;
      float float_j2 = (float)(j + 1);

      g1 *= g1;
      g2 *= g2;

      g1 *= nrcp;
      g2 *= nrcp;
	
      cconst.x = float_j1*(1.0f/32768.0f);
      cconst.y = float_j2*(1.0f/32768.0f);

      multiplier1 = 1.0f;
      g1 = native_exp(g1);

      multiplier2 = 1.0f;
      g2 = native_exp(g2);

      multiplier1 -= g1;
      multiplier2 -= g2;

      multiplier1 = native_sqrt(multiplier1);
      multiplier2 = native_sqrt(multiplier2);

      cconst.z = multiplier1 * normalize_val;
      cconst.w = multiplier2 * normalize_val;
    }
  else
    {
      g1 = (float)(16383 -(j - 16384)) + 0.5f;
      int j1 = j - 16384;
      g2 = (float)(16383 -(j - 16384 + 1)) + 0.5f;
      int j2 = j - 16384 + 1;
		
      g1 *= g1;
      g2 *= g2;

      g1 *= nrcp;
      g2 *= nrcp;
	
      cconst.x = (float)(j1 + 16384)*(1.0f/32768.0f) -1.0f;
      cconst.y = (float)(j2 + 16384)*(1.0f/32768.0f) -1.0f;
	
      multiplier1 = 1.0f;
      g1 = native_exp(g1);

      multiplier2 = 1.0f;
      g2 = native_exp(g2);

      multiplier1 -= g1;
      multiplier2 -= g2;

      multiplier1 = native_sqrt(multiplier1);
      multiplier2 = native_sqrt(multiplier2);

      cconst.z = multiplier1 * normalize_val;
      cconst.w = multiplier2 * normalize_val;
    }

	//R: each work item will process 2 complex data elements and write 2*2*16 new complex elements into dechirped array
	
	float4 data=gpu_data[tid+dchunk*(FFT_SIZE/2)];
	float4 cur_chirp1;
	float4 cur_chirp2;
	float4 cur_chirp3;
	float4 cur_chirp4;
	float4 cur_dechirp11, cur_dechirp12;
	float4 cur_dechirp21, cur_dechirp22;
	float4 cur_dechirp31, cur_dechirp32;
	float4 cur_dechirp41, cur_dechirp42;
	float xx, yy, zz, ww, yx, xy, wz, zw;
	float xx2, yy2, zz2, ww2, yx2, xy2, wz2, zw2;
	float xx3, yy3, zz3, ww3, yx3, xy3, wz3, zw3;
	float xx4, yy4, zz4, ww4, yx4, xy4, wz4, zw4;
	
	for(uint i = 0; i < 16; i += 4)
	  {    //R: can be optimized via mad instruction probably

    		cur_chirp1 = calc_chirp(cconst, dm_start, i);
    		cur_chirp2 = calc_chirp(cconst, dm_start, i+1);
    		cur_chirp3 = calc_chirp(cconst, dm_start, i+2);
    		cur_chirp4 = calc_chirp(cconst, dm_start, i+3);

		//negative sign
		yy = data.y*cur_chirp1.y;
		xy = data.x*cur_chirp1.y;
		ww = data.w*cur_chirp1.w;
		zw = data.z*cur_chirp1.w;
		yy2 = data.y*cur_chirp2.y;
		xy2 = data.x*cur_chirp2.y;
		ww2 = data.w*cur_chirp2.w;
		zw2 = data.z*cur_chirp2.w;
		yy3 = data.y*cur_chirp3.y;
		xy3 = data.x*cur_chirp3.y;
		ww3 = data.w*cur_chirp3.w;
		zw3 = data.z*cur_chirp3.w;
		yy4 = data.y*cur_chirp4.y;
		xy4 = data.x*cur_chirp4.y;
		ww4 = data.w*cur_chirp4.w;
		zw4 = data.z*cur_chirp4.w;

		cur_dechirp11.x = mad(data.x, cur_chirp1.x,  yy);
		cur_dechirp12.y = mad(data.y, cur_chirp1.x,  xy);
		cur_dechirp11.z = mad(data.z, cur_chirp1.z,  ww);
		cur_dechirp12.w = mad(data.w, cur_chirp1.z,  zw);
		cur_dechirp21.x = mad(data.x, cur_chirp2.x, yy2);
		cur_dechirp22.y = mad(data.y, cur_chirp2.x, xy2);
		cur_dechirp21.z = mad(data.z, cur_chirp2.z, ww2);
		cur_dechirp22.w = mad(data.w, cur_chirp2.z, zw2);
		cur_dechirp31.x = mad(data.x, cur_chirp3.x, yy3);
		cur_dechirp32.y = mad(data.y, cur_chirp3.x, xy3);
		cur_dechirp31.z = mad(data.z, cur_chirp3.z, ww3);
		cur_dechirp32.w = mad(data.w, cur_chirp3.z, zw3);
		cur_dechirp41.x = mad(data.x, cur_chirp4.x, yy4);
		cur_dechirp42.y = mad(data.y, cur_chirp4.x, xy4);
		cur_dechirp41.z = mad(data.z, cur_chirp4.z, ww4);
		cur_dechirp42.w = mad(data.w, cur_chirp4.z, zw4);

/*		yy = -yy;
		xy = -xy;
		ww = -ww;
		zw = -zw;
		yy2 = -yy2;
		xy2 = -xy2;
		ww2 = -ww2;
		zw2 = -zw2;
		yy3 = -yy3;
		xy3 = -xy3;
		ww3 = -ww3;
		zw3 = -zw3;
		yy4 = -yy4;
		xy4 = -xy4;
		ww4 = -ww4;
		zw4 = -zw4;
*/		
		cur_dechirp12.x = data.x * cur_chirp1.x - yy ;
		cur_dechirp11.y = data.y * cur_chirp1.x - xy ;
		cur_dechirp12.z = data.z * cur_chirp1.z - ww ;
		cur_dechirp11.w = data.w * cur_chirp1.z - zw ;
		cur_dechirp22.x = data.x * cur_chirp2.x - yy2;
		cur_dechirp21.y = data.y * cur_chirp2.x - xy2;
		cur_dechirp22.z = data.z * cur_chirp2.z - ww2;
		cur_dechirp21.w = data.w * cur_chirp2.z - zw2;
		cur_dechirp32.x = data.x * cur_chirp3.x - yy3;
		cur_dechirp31.y = data.y * cur_chirp3.x - xy3;
		cur_dechirp32.z = data.z * cur_chirp3.z - ww3;
		cur_dechirp31.w = data.w * cur_chirp3.z - zw3;
		cur_dechirp42.x = data.x * cur_chirp4.x - yy4;
		cur_dechirp41.y = data.y * cur_chirp4.x - xy4;
		cur_dechirp42.z = data.z * cur_chirp4.z - ww4;
		cur_dechirp41.w = data.w * cur_chirp4.z - zw4;
/*
		cur_dechirp12.x = mad(data.x, cur_chirp1.x, yy);
		cur_dechirp11.y = mad(data.y, cur_chirp1.x, xy);
		cur_dechirp12.z = mad(data.z, cur_chirp1.z, ww);
		cur_dechirp11.w = mad(data.w, cur_chirp1.z, zw);
		cur_dechirp22.x = mad(data.x, cur_chirp2.x, yy2);
		cur_dechirp21.y = mad(data.y, cur_chirp2.x, xy2);
		cur_dechirp22.z = mad(data.z, cur_chirp2.z, ww2);
		cur_dechirp21.w = mad(data.w, cur_chirp2.z, zw2);
		cur_dechirp32.x = mad(data.x, cur_chirp3.x, yy3);
		cur_dechirp31.y = mad(data.y, cur_chirp3.x, xy3);
		cur_dechirp32.z = mad(data.z, cur_chirp3.z, ww3);
		cur_dechirp31.w = mad(data.w, cur_chirp3.z, zw3);
		cur_dechirp42.x = mad(data.x, cur_chirp4.x, yy4);
		cur_dechirp41.y = mad(data.y, cur_chirp4.x, xy4);
		cur_dechirp42.z = mad(data.z, cur_chirp4.z, ww4);
		cur_dechirp41.w = mad(data.w, cur_chirp4.z, zw4);
*/
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+0)*(FFT_SIZE/2)+tid] = cur_dechirp11;
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+1)*(FFT_SIZE/2)+tid] = cur_dechirp12;
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+2)*(FFT_SIZE/2)+tid] = cur_dechirp21;
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+3)*(FFT_SIZE/2)+tid] = cur_dechirp22;
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+4)*(FFT_SIZE/2)+tid] = cur_dechirp31;
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+5)*(FFT_SIZE/2)+tid] = cur_dechirp32;
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+6)*(FFT_SIZE/2)+tid] = cur_dechirp41;
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+7)*(FFT_SIZE/2)+tid] = cur_dechirp42;

		}
}

#else
__kernel void dechirp_range1_kernel ( __global float4* gpu_data ,
										__global float4* gpu_dechirped,
										__const float dm_start)
{

	uint tid = get_global_id(0);
	uint dchunk=get_global_id(1);
	//R: each work item will process 2 complex data elements and write 2*2*16 new complex elements into dechirped array
	
	float4 data=gpu_data[tid+dchunk*(FFT_SIZE/2)];
	float4 cur_chirp;
	float4 cur_dechirp;
	float xx, yy, zz, ww, yx, xy, wz, zw;
	
	for(uint i=0;i<16;i++){//R: can be optimized via mad instruction probably
		
		cur_chirp = calc_chirp(tid *2, i, dm_start);
		
		//negative sign
		xx = data.x*cur_chirp.x;
		yy = data.y*cur_chirp.y;
		yx = data.y*cur_chirp.x;
		xy = data.x*cur_chirp.y;
		zz = data.z*cur_chirp.z;
		ww = data.w*cur_chirp.w;
		wz = data.w*cur_chirp.z;
		zw = data.z*cur_chirp.w;
#if APV505
		cur_dechirp.x = xx - yy;
		cur_dechirp.y = yx + xy;
		cur_dechirp.z = zz - ww;
		cur_dechirp.w = wz + zw;
#else
		cur_dechirp.x = xx + yy;
		cur_dechirp.y = yx - xy;
		cur_dechirp.z = zz + ww;
		cur_dechirp.w = wz - zw;
#endif
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+0)*(FFT_SIZE/2)+tid]=cur_dechirp;

		//positive sign
#if APV505
		cur_dechirp.x = xx + yy;
		cur_dechirp.y = yx - xy;
		cur_dechirp.z = zz + ww;
		cur_dechirp.w = wz - zw;
#else
		cur_dechirp.x = xx - yy;
		cur_dechirp.y = yx + xy;
		cur_dechirp.z = zz - ww;
		cur_dechirp.w = wz + zw;
#endif
		gpu_dechirped[32*(FFT_SIZE/2)*dchunk+(2*i+1)*(FFT_SIZE/2)+tid]=cur_dechirp;

		}
}
#endif
#endif
#else	// dechirping via look up table

//R: dechirps whole dm from initial single gpu_data array
//data packed in float4 a:
//a.s0<->data[0].re;a.s1<->data[0].im;a.s1<->data[1].re and so on
//in dechirped minus sign goes first, next is positive sign, next is negative for next DM.
__kernel void dechirp_range1_kernel(__global float4* gpu_data ,
                             __global float4* gpu_chirps,
                             __global float4* gpu_dechirped)
{
	uint tid = get_global_id(0);
	uint dchunk=get_global_id(1);
	//R: each work item will process 2 complex data elements and write 2*2*16 new complex elements into dechirped array
	float4 data=gpu_data[tid+dchunk*(32768/2)];
	float4 cur_chirp;
	float4 cur_dechirp;
	for(uint i=0;i<16;i++){//R: can be optimized via mad instruction probably
		cur_chirp=gpu_chirps[i*(32768/2)+tid];
		//negative sign
#if APV505
		cur_dechirp.x = data.x*cur_chirp.x - data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x + data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z - data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z + data.z*cur_chirp.w;
#else
		cur_dechirp.x = data.x*cur_chirp.x + data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x - data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z + data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z - data.z*cur_chirp.w;
#endif
		gpu_dechirped[32*(32768/2)*dchunk+(2*i+0)*(32768/2)+tid]=cur_dechirp;
		//positive sign
#if APV505
		cur_dechirp.x = data.x*cur_chirp.x + data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x - data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z + data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z - data.z*cur_chirp.w;
#else
		cur_dechirp.x = data.x*cur_chirp.x - data.y*cur_chirp.y;
		cur_dechirp.y = data.y*cur_chirp.x + data.x*cur_chirp.y;
		cur_dechirp.z = data.z*cur_chirp.z - data.w*cur_chirp.w;
		cur_dechirp.w = data.w*cur_chirp.z + data.z*cur_chirp.w;
#endif
		gpu_dechirped[32*(32768/2)*dchunk+(2*i+1)*(32768/2)+tid]=cur_dechirp;
	}
}

#endif // end of combined dechirp kernel if/else-swith


__kernel void half_temp_range1_kernel(__global float8* gpu_dechirped, __global float4* gpu_power,const float ravg_pow){
	uint tid = get_global_id(0);
	uint dchunk=get_global_id(1);
	uint i=get_global_id(2);
	
	float8 temp;
	float4 power;

	temp = gpu_dechirped[ (dchunk<<18) + (i<<13) + tid ];

	power.xy =  mad( temp.s02, temp.s02, (temp.s13 * temp.s13) ) * ravg_pow;
		
	power.zw =  mad( temp.s46 ,temp.s46, (temp.s57 * temp.s57) ) * ravg_pow;
		
	gpu_power[ (dchunk<<17) + (i<<12) + tid ] = power;
	
}
#if 0
__kernel void PC_single_pulse_kernel_FFA_update_vectorised(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,const int need_large_FFA_data,
									   __global float4* small_neg,__global float4* small_pos,
									   __global float4* large_neg,__global float4* large_pos)
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint len4 = 32768 >> 3;
	uint4 coadd_offsets=(uint4)(0);
	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshld for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshld for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshld for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshld for vector operations
	uint result=0;//R: will contain info about location of found pulses and best signals
	int4 pulse0=(int4)(0);//R: result of comparison
	int4 pulse1=(int4)(0);//R: result of comparison
	int4 pulse2=(int4)(0);//R: result of comparison
	int4 pulse3=(int4)(0);//R: result of comparison
	uint4 was_pulse=(uint4)(0);
	//R: low 16 bits for pulses above threshold, each coadd level coded as enabled bit in corresponding position
	//R: high 16 bits for best pulses, if were, again, enabled bit means found pulse

//R: will do unroll for coadds to save reads from global memory
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	coadd_begins.y=num_dchunks*32*(32768>>3);
	coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	for(int m=0;m<len4;m+=16){//R:loading data piece of 16*4 samples
		int m1=coadd_offsets.x+m;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];

			pulse0|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in memory
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: next coadd level ready to check for pulses
			pulse1|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			pulse2|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			pulse3|=(d0>tt3)|(d1>tt3);
			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=(m>>4)+next_offset;
			gpu_power[m1]=d0;
			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+(m>>4)+dchunk*(32768>>7)]=d0;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+(m>>4)+dchunk*(32768>>7)]=d0;
				}
			}
	}//R: initial and 3 coadded arrays are checked, now need to store results of check and reload for next levels
	was_pulse.x=1&(pulse0.x|pulse0.y|pulse0.z|pulse0.w);
	was_pulse.y=1&(pulse1.x|pulse1.y|pulse1.z|pulse1.w);
	was_pulse.z=1&(pulse2.x|pulse2.y|pulse2.z|pulse2.w);
	was_pulse.w=1&(pulse3.x|pulse3.y|pulse3.z|pulse3.w);

	result+=(was_pulse.x)+(was_pulse.y<<1)+(was_pulse.z<<2)+(was_pulse.w<<3);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	t=thresh[1];
	tt0=(float4)(t.x);
	tt1=(float4)(t.y);
	tt2=(float4)(t.z);
	tt3=(float4)(t.w);
	was_pulse=(uint4)(0);
	pulse0=(int4)(0);
	pulse1=(int4)(0);
	pulse2=(int4)(0);
	pulse3=(int4)(0);

	len4 = 32768 >>(3+4);//R:4 coadds were done already
	coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>7)*32;
	coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>8)*32;
	coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>9)*32;
	next_begin=coadd_begins.w+num_dchunks*(32768>>10)*32;
	coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>7);
	coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>8);
	coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>9);
	coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>10);
	next_offset=next_begin+(dchunk*32+tid)*(32768>>11);
//R:now repeating same loop as for first
	for(int m=0;m<len4;m+=16){//R:loading data piece of 16*4 samples
		int m1=m+coadd_offsets.x;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 4 check
			pulse0|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in memory
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 5 check
			pulse1|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: next coadd level starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: coadd level 6 check
			pulse2|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)]=d0;
					large_neg[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)+1]=d1;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)]=d0;
					large_pos[need_large_FFA_data+(m>>3)+dchunk*(32768>>10)+1]=d1;
				}
			}

			//R: coadd level 7 check
			pulse3|=(d0>tt3)|(d1>tt3);
			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=(m>>4)+next_offset;
			gpu_power[m1]=d0;//R: coadd level 8 store
	}
	was_pulse.x=1&(pulse0.x|pulse0.y|pulse0.z|pulse0.w);
	was_pulse.y=1&(pulse1.x|pulse1.y|pulse1.z|pulse1.w);
	was_pulse.z=1&(pulse2.x|pulse2.y|pulse2.z|pulse2.w);
	was_pulse.w=1&(pulse3.x|pulse3.y|pulse3.z|pulse3.w);
	result+=(was_pulse.x<<4)+(was_pulse.y<<5)+(was_pulse.z<<6)+(was_pulse.w<<7);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	t=thresh[2];
	tt0=(float4)(t.x);
	tt1=(float4)(t.y);
	tt2=(float4)(t.z);
	tt3=(float4)(t.w);
	was_pulse=(uint4)(0);
	pulse0=(int4)(0);
	pulse1=(int4)(0);
	pulse2=(int4)(0);
	pulse3=(int4)(0);
	coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>11)*32;
	coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>12)*32;
	coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>13)*32;
	next_begin=coadd_begins.w+num_dchunks*(32768>>14)*32;
	coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>11);
	coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>12);
	coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>13);
	coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>14);
	next_offset=next_begin+(dchunk*32+tid)*(32768>>15);
//R: 32768>>11==16 so we have only 16 float4 elements, no loop by m needed.
		int m1=coadd_offsets.x;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 8 check
			pulse0|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in memory
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 9 check,last needed coadd
			pulse1|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
	was_pulse.x=1&(pulse0.x|pulse0.y|pulse0.z|pulse0.w);
	was_pulse.y=1&(pulse1.x|pulse1.y|pulse1.z|pulse1.w);
	result+=(was_pulse.x<<8)+(was_pulse.y<<9);
//R: last we do is to write result of pulse finding
	results[32*dchunk+tid]=result;
}

#endif
//R: single pulse is rare event. That is, buffer readback is rare too. So, no need to care about what part of buffer
//should be read back to host memory (and current implementation will read full buffer anyway).
//This allows to save on flags (single flag yes/no needed instead of detailed info where was pulse.
//Also, this allows to split kernel on many just as I did with pulse finding in MB. If one of parts finds pulse
//and corresponding FFA updates not needed, other parts can be skipped completely.
//Splitting kernel will expose greater parallelism and will do much more effective memory latence hiding

//reduce0 will update only small FFA buffers, higher coadds go in next kernel calls

#if PSFIND0 || PSFIND1 //petri33's SingleFind kernel modification

#pragma OPENCL EXTENSION cl_khr_byte_addressable_store: enable

//
// petri33 : Greater Than using CUDA Predicates (hence gtp)
// helper functions (inlined automatically)
//
typedef struct {bool x; bool y;} bool2;

bool2 or2(bool2 a, bool2 b)
{
  bool2 ret;

  ret.x = a.x | b.x;
  ret.y = a.y | b.y;

  return ret;
}


bool or1(bool2 a, bool2 b)
{
  bool2 r = or2(a, b);

  return r.x | r.y;
}


bool2 gtp(float4 a, float4 cc) 
{
  bool p1 = (a.x > cc.x);
  bool p2 = (a.y > cc.y);
  bool p3 = (a.z > cc.z);
  bool p4 = (a.w > cc.w);

  p1 |= p3;
  p2 |= p4;

  bool2 ret;

  ret.x = p1;
  ret.y = p2;

  return ret;
}


bool2 gtp2(float4 a, float4 b, float4 cc) 
{
  bool2 r1 = gtp(a, cc);
  bool2 r2 = gtp(b, cc);
  return or2(r1, r2);
}


bool2 gtp4(float4 a, float4 b, float4 c, float4 d, float4 cc)
{
  bool2 r1 = gtp2(a, b, cc);
  bool2 r2 = gtp2(c, d, cc);
  return or2(r1, r2);
}


bool2 gtp8(float4 a, float4 b, float4 c, float4 d, float4 a2, float4 b2, float4 c2, float4 d2, float4 cc)
{
  bool2 r1 = gtp4(a, b, c, d, cc);
  bool2 r2 = gtp4(a2, b2, c2, d2, cc);
  return or2(r1, r2);
}
#endif
#if 1

#if PSFIND0
//
// petri33:
// modified code for big or blocks. More instruction level parallelism for Kepler and upwards. 
// Possibly good for other architectures.
//
__kernel __attribute__((reqd_work_group_size(1, 1, 64)))
void PC_single_pulse_kernel_FFA_update_reduce0(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,//const int need_large_FFA_data,
									   __global float4* small_neg,__global float4* small_pos//,
									   //__global float4* large_neg,__global float4* large_pos
									   )
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	int mchunk=get_global_id(2);
//	uint len4 = 32768 >> 3;
//	uint4 coadd_offsets=(uint4)(0);
//	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	//uint result=0;
	//int4 pulse0=(int4)(0);//R: result of comparison
	//int4 pulse1=(int4)(0);//R: result of comparison
	//int4 pulse2=(int4)(0);//R: result of comparison
	//int4 pulse3=(int4)(0);//R: result of comparison
	bool2 was_pulse1, was_pulse2, was_pulse3, was_pulse4;
	//int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	//R: Coadded values storage address arithmetics now greatly simplified
	//R: only 1 coadd level stored per kernel call so max 2 coadds need to be addressed
	next_begin=num_dchunks*32*(32768>>3);//R: area just after initial power array
	next_offset=next_begin+(dchunk*32+tid)*(32768>>7);//R: don't account for mchunk offset

	//coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	//coadd_begins.y=num_dchunks*32*(32768>>3);
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(dchunk*32+tid)*(32768>>3)+(mchunk<<4);//R:reading corresponding mchunk (16*4 samples here) from power array - coadd level 0
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];

			was_pulse1 = gtp4(d0,  d1,  d2,  d3,  tt0);
			was_pulse2 = gtp4(d4,  d5,  d6,  d7,  tt0); 
                        was_pulse3 = gtp4(d8,  d9,  d10, d11, tt0);
			was_pulse4 = gtp4(d12, d13, d14, d15, tt0);
			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);
			was_pulse1 = or2(was_pulse1, was_pulse2);

			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;

			//R: next coadd level ready to check for pulses
			was_pulse2 = gtp4(d0, d1, d2, d3, tt1);
			was_pulse3 = gtp4(d4, d5, d6, d7, tt1);
			was_pulse2 = or2(was_pulse2, was_pulse3);

			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			was_pulse3 = gtp4(d0, d1, d2, d3, tt2);

			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			was_pulse4 = gtp2(d0, d1, tt3);

			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=mchunk+next_offset;
			gpu_power[m1]=d0;
			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);

			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}
			}
	}
	bool was_pulse = or1(was_pulse1, was_pulse2);
	//was_pulse.x=was_pulse.x||was_pulse.y||was_pulse.z||was_pulse.w;//result reduction
	if(was_pulse) results[0]=1;
}
__kernel __attribute__((reqd_work_group_size(1, 1, 64)))
void PC_single_pulse_kernel_FFA_update_reduce0_BLANKIT(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,//const int need_large_FFA_data,
									   __global float4* small_neg,__global float4* small_pos, __constant uint* need_blanking//,
									   //__global float4* large_neg,__global float4* large_pos
									   )
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	int mchunk=get_global_id(2);

	if(need_blanking[dchunk]){
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=16.0f;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=16.0f;
				}
			}
			return;
	}

//	uint len4 = 32768 >> 3;
//	uint4 coadd_offsets=(uint4)(0);
//	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	//uint result=0;
	//int4 pulse0=(int4)(0);//R: result of comparison
	//int4 pulse1=(int4)(0);//R: result of comparison
	//int4 pulse2=(int4)(0);//R: result of comparison
	//int4 pulse3=(int4)(0);//R: result of comparison
	bool2 was_pulse1, was_pulse2, was_pulse3, was_pulse4;
	//int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	//R: Coadded values storage address arithmetics now greatly simplified
	//R: only 1 coadd level stored per kernel call so max 2 coadds need to be addressed
	next_begin=num_dchunks*32*(32768>>3);//R: area just after initial power array
	next_offset=next_begin+(dchunk*32+tid)*(32768>>7);//R: don't account for mchunk offset

	//coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	//coadd_begins.y=num_dchunks*32*(32768>>3);
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(dchunk*32+tid)*(32768>>3)+(mchunk<<4);//R:reading corresponding mchunk (16*4 samples here) from power array - coadd level 0
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];

			was_pulse1 = gtp4(d0,  d1,  d2,  d3,  tt0);
			was_pulse2 = gtp4(d4,  d5,  d6,  d7,  tt0); 
                        was_pulse3 = gtp4(d8,  d9,  d10, d11, tt0);
			was_pulse4 = gtp4(d12, d13, d14, d15, tt0);
			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);
			was_pulse1 = or2(was_pulse1, was_pulse2);

			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;

			//R: next coadd level ready to check for pulses
			was_pulse2 = gtp4(d0, d1, d2, d3, tt1);
			was_pulse3 = gtp4(d4, d5, d6, d7, tt1);
			was_pulse2 = or2(was_pulse2, was_pulse3);

			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			was_pulse3 = gtp4(d0, d1, d2, d3, tt2);

			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			was_pulse4 = gtp2(d0, d1, tt3);

			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=mchunk+next_offset;
			gpu_power[m1]=d0;
			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);

			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}
			}
	}
	bool was_pulse = or1(was_pulse1, was_pulse2);
	//was_pulse.x=was_pulse.x||was_pulse.y||was_pulse.z||was_pulse.w;//result reduction
	if(was_pulse) results[0]=1;
}

#else
__kernel __attribute__((reqd_work_group_size(1, 1, 64)))
void PC_single_pulse_kernel_FFA_update_reduce0(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,//const int need_large_FFA_data,
									   __global float4* small_neg,__global float4* small_pos//,
									   //__global float4* large_neg,__global float4* large_pos
									   )
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	int mchunk=get_global_id(2);
//	uint len4 = 32768 >> 3;
//	uint4 coadd_offsets=(uint4)(0);
//	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	//uint result=0;
	//int4 pulse0=(int4)(0);//R: result of comparison
	//int4 pulse1=(int4)(0);//R: result of comparison
	//int4 pulse2=(int4)(0);//R: result of comparison
	//int4 pulse3=(int4)(0);//R: result of comparison
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	//R: Coadded values storage address arithmetics now greatly simplified
	//R: only 1 coadd level stored per kernel call so max 2 coadds need to be addressed
	next_begin=num_dchunks*32*(32768>>3);//R: area just after initial power array
	next_offset=next_begin+(dchunk*32+tid)*(32768>>7);//R: don't account for mchunk offset

	//coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	//coadd_begins.y=num_dchunks*32*(32768>>3);
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(dchunk*32+tid)*(32768>>3)+(mchunk<<4);//R:reading corresponding mchunk (16*4 samples here) from power array - coadd level 0
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];

			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: next coadd level ready to check for pulses
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			was_pulse|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			was_pulse|=(d0>tt3)|(d1>tt3);
			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=mchunk+next_offset;
			gpu_power[m1]=d0;
			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}
			}
	}
	was_pulse.x=was_pulse.x||was_pulse.y||was_pulse.z||was_pulse.w;//result reduction
	if(was_pulse.x) results[0]=1;
}
__kernel __attribute__((reqd_work_group_size(1, 1, 64)))
void PC_single_pulse_kernel_FFA_update_reduce0_BLANKIT(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,//const int need_large_FFA_data,
									   __global float4* small_neg,__global float4* small_pos, __constant uint* need_blanking//,
									   //__global float4* large_neg,__global float4* large_pos
									   )
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	int mchunk=get_global_id(2);

	if(need_blanking[dchunk]){
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=16.0f;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=16.0f;
				}
			}
			return;
	}

//	uint len4 = 32768 >> 3;
//	uint4 coadd_offsets=(uint4)(0);
//	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	//uint result=0;
	//int4 pulse0=(int4)(0);//R: result of comparison
	//int4 pulse1=(int4)(0);//R: result of comparison
	//int4 pulse2=(int4)(0);//R: result of comparison
	//int4 pulse3=(int4)(0);//R: result of comparison
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	//R: Coadded values storage address arithmetics now greatly simplified
	//R: only 1 coadd level stored per kernel call so max 2 coadds need to be addressed
	next_begin=num_dchunks*32*(32768>>3);//R: area just after initial power array
	next_offset=next_begin+(dchunk*32+tid)*(32768>>7);//R: don't account for mchunk offset

	//coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	//coadd_begins.y=num_dchunks*32*(32768>>3);
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(dchunk*32+tid)*(32768>>3)+(mchunk<<4);//R:reading corresponding mchunk (16*4 samples here) from power array - coadd level 0
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];

			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: next coadd level ready to check for pulses
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			was_pulse|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			was_pulse|=(d0>tt3)|(d1>tt3);
			//R: now only coadd and store, need to refresh array in registers so memory read required
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			m1=mchunk+next_offset;
			gpu_power[m1]=d0;
			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){
				//R: Store negative DM sign power data for small FFA
					small_neg[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[need_small_FFA_data+mchunk+dchunk*(32768>>7)]=d0;
				}
			}
	}
	was_pulse.x=was_pulse.x||was_pulse.y||was_pulse.z||was_pulse.w;//result reduction
	if(was_pulse.x) results[0]=1;
}
#endif

#if PSFIND1
__kernel __attribute__((reqd_work_group_size(32, 1, 2)))
void PC_single_pulse_kernel_FFA_update_reduce1(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   //const int need_small_FFA_data,
									   const int need_large_FFA_data,
									   //__global float4* small_neg,__global float4* small_pos//,
									   __global float4* large_neg,__global float4* large_pos
									   )
{
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	float4 t=thresh[1];
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	t=thresh[2];//for last 2 levels
	bool2 was_pulse1, was_pulse2, was_pulse3, was_pulse4; //
	//int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
	int initial_begin=num_dchunks*32*(32768>>3);
	int initial_offset=initial_begin+(dchunk*32+tid)*(32768>>7);//R:we reading already coadded data here, for particular dchunk
	//int next_offset=initial_begin+num_dchunks*(32768>>7)*32+(dchunk*32+tid)*(32768>>11);
	//len4 = 32768 >>(3+4);//R:4 coadds were done already

	//coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	//coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>7)*32;
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>8)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>9)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>10)*32;
	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>7);
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>8);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>9);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>10);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>11);
//R:now repeating same loop as for first
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(mchunk<<4)+initial_offset;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 4 check
			was_pulse1 = gtp4(d0, d1, d2, d3, tt0);
			was_pulse2 = gtp4(d4, d5, d6, d7, tt0);
                        was_pulse3 = gtp4(d8, d9, d10, d11, tt0);
			was_pulse4 = gtp4(d12, d13, d14, d15, tt0);
			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);
			was_pulse1 = or2(was_pulse1, was_pulse2);

			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 5 check
			was_pulse2 = gtp4(d0, d1, d2, d3, tt1);
			was_pulse3 = gtp4(d4, d5, d6, d7, tt1);
			was_pulse2 = or2(was_pulse2, was_pulse3);

			//R: next coadd level starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: coadd level 6 check
			was_pulse3 = gtp4(d0, d1, d2, d3, tt2);
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}
			}

			//R: coadd level 7 check
			was_pulse4 = gtp2(d0, d1, tt3);
			//was_pulse.x |= (was_pulse.y||was_pulse.z||was_pulse.w);//result reduction
			d0.xy = d0.xz + d0.yw; d0.zw = d1.xz + d1.yw;
			//R: coadd level 8 check
			was_pulse3 = or2(was_pulse3, was_pulse4);
			bool2 wp1;
			wp1.x = (d0.x > t.x);
			wp1.y = (d0.y > t.x);
			bool2 wp2;
			wp2.x = (d0.z > t.x);
			wp2.y = (d0.z > t.x);
			
			was_pulse4 = or2(wp1, wp2);

			//R: now do coadd and save coadded piece in registers
			d0.xy = d0.xz + d0.yw;
			//R: coadd level 9 check,last needed coadd

			bool2 wp3;
			wp3.x = (d0.x > t.y);
			wp3.y = (d0.y > t.y);

			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);
			was_pulse1 = or2(was_pulse1, wp3);
	}
	bool was_pulse = or1(was_pulse1, was_pulse2);
	if(was_pulse) results[0]=1;
}
__kernel __attribute__((reqd_work_group_size(32, 1, 2)))
void PC_single_pulse_kernel_FFA_update_reduce1_BLANKIT(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   //const int need_small_FFA_data,
									   const int need_large_FFA_data,
									   //__global float4* small_neg,__global float4* small_pos//,
									   __global float4* large_neg,__global float4* large_pos, __constant uint* need_blanking
									   )
{
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);

	if(need_blanking[dchunk]){
		//printf("Blanking apllied dchunk=%d\n",dchunk);
		if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=128.f;
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=128.f;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=128.f;
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=128.f;
				}
		}
		return;
	}
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	float4 t=thresh[1];
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	t=thresh[2];//for last 2 levels
	bool2 was_pulse1, was_pulse2, was_pulse3, was_pulse4; //
	//int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
	int initial_begin=num_dchunks*32*(32768>>3);
	int initial_offset=initial_begin+(dchunk*32+tid)*(32768>>7);//R:we reading already coadded data here, for particular dchunk
	//int next_offset=initial_begin+num_dchunks*(32768>>7)*32+(dchunk*32+tid)*(32768>>11);
	//len4 = 32768 >>(3+4);//R:4 coadds were done already

	//coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	//coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>7)*32;
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>8)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>9)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>10)*32;
	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>7);
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>8);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>9);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>10);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>11);
//R:now repeating same loop as for first
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(mchunk<<4)+initial_offset;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 4 check
			was_pulse1 = gtp4(d0, d1, d2, d3, tt0);
			was_pulse2 = gtp4(d4, d5, d6, d7, tt0);
                        was_pulse3 = gtp4(d8, d9, d10, d11, tt0);
			was_pulse4 = gtp4(d12, d13, d14, d15, tt0);
			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);
			was_pulse1 = or2(was_pulse1, was_pulse2);

			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 5 check
			was_pulse2 = gtp4(d0, d1, d2, d3, tt1);
			was_pulse3 = gtp4(d4, d5, d6, d7, tt1);
			was_pulse2 = or2(was_pulse2, was_pulse3);

			//R: next coadd level starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: coadd level 6 check
			was_pulse3 = gtp4(d0, d1, d2, d3, tt2);
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}
			}

			//R: coadd level 7 check
			was_pulse4 = gtp2(d0, d1, tt3);
			//was_pulse.x |= (was_pulse.y||was_pulse.z||was_pulse.w);//result reduction
			d0.xy = d0.xz + d0.yw; d0.zw = d1.xz + d1.yw;
			//R: coadd level 8 check
			was_pulse3 = or2(was_pulse3, was_pulse4);
			bool2 wp1;
			wp1.x = (d0.x > t.x);
			wp1.y = (d0.y > t.x);
			bool2 wp2;
			wp2.x = (d0.z > t.x);
			wp2.y = (d0.z > t.x);
			
			was_pulse4 = or2(wp1, wp2);

			//R: now do coadd and save coadded piece in registers
			d0.xy = d0.xz + d0.yw;
			//R: coadd level 9 check,last needed coadd

			bool2 wp3;
			wp3.x = (d0.x > t.y);
			wp3.y = (d0.y > t.y);

			was_pulse1 = or2(was_pulse1, was_pulse2);
			was_pulse2 = or2(was_pulse3, was_pulse4);
			was_pulse1 = or2(was_pulse1, wp3);
	}
	bool was_pulse = or1(was_pulse1, was_pulse2);
	if(was_pulse) results[0]=1;
}
// modified code ends
#else
__kernel __attribute__((reqd_work_group_size(32, 1, 2)))
void PC_single_pulse_kernel_FFA_update_reduce1(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   //const int need_small_FFA_data,
									   const int need_large_FFA_data,
									   //__global float4* small_neg,__global float4* small_pos//,
									   __global float4* large_neg,__global float4* large_pos
									   )
{
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	float4 t=thresh[1];
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	t=thresh[2];//for last 2 levels
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
	int initial_begin=num_dchunks*32*(32768>>3);
	int initial_offset=initial_begin+(dchunk*32+tid)*(32768>>7);//R:we reading already coadded data here, for particular dchunk
	//int next_offset=initial_begin+num_dchunks*(32768>>7)*32+(dchunk*32+tid)*(32768>>11);
	//len4 = 32768 >>(3+4);//R:4 coadds were done already

	//coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	//coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>7)*32;
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>8)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>9)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>10)*32;
	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>7);
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>8);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>9);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>10);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>11);
//R:now repeating same loop as for first
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(mchunk<<4)+initial_offset;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 4 check
			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 5 check
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: next coadd level starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: coadd level 6 check
			was_pulse|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}
			}

			//R: coadd level 7 check
			was_pulse|=(d0>tt3)|(d1>tt3);
			was_pulse.x|=was_pulse.y|was_pulse.z|was_pulse.w;//result reduction
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			//R: coadd level 8 check
			was_pulse.x|=(d0.x>t.x)|(d0.y>t.x)|(d0.z>t.x)|(d0.w>t.x);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;
			//R: coadd level 9 check,last needed coadd
			was_pulse.x|=(d0.x>t.y)|(d0.y>t.y);
	}
	if(was_pulse.x) results[0]=1;
}
__kernel __attribute__((reqd_work_group_size(32, 1, 2)))
void PC_single_pulse_kernel_FFA_update_reduce1_BLANKIT(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   //const int need_small_FFA_data,
									   const int need_large_FFA_data,
									   //__global float4* small_neg,__global float4* small_pos//,
									   __global float4* large_neg,__global float4* large_pos, __constant uint* need_blanking
									   )
{
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);

	if(need_blanking[dchunk]){
		//printf("Blanking apllied dchunk=%d\n",dchunk);
		if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=128.f;
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=128.f;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=128.f;
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=128.f;
				}
		}
		return;
	}
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
	float4 t=thresh[1];
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	t=thresh[2];//for last 2 levels
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;
	int initial_begin=num_dchunks*32*(32768>>3);
	int initial_offset=initial_begin+(dchunk*32+tid)*(32768>>7);//R:we reading already coadded data here, for particular dchunk
	//int next_offset=initial_begin+num_dchunks*(32768>>7)*32+(dchunk*32+tid)*(32768>>11);
	//len4 = 32768 >>(3+4);//R:4 coadds were done already

	//coadd_begins.x=next_begin;//R: will fetch data stored at the end of prev coadd bunch loop
	//coadd_begins.y=coadd_begins.x+num_dchunks*(32768>>7)*32;
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>8)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>9)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>10)*32;
	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>7);
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>8);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>9);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>10);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>11);
//R:now repeating same loop as for first
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=(mchunk<<4)+initial_offset;
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];
			d8=gpu_power[m1+8];d9=gpu_power[m1+9];d10=gpu_power[m1+10];d11=gpu_power[m1+11];
			d12=gpu_power[m1+12];d13=gpu_power[m1+13];d14=gpu_power[m1+14];d15=gpu_power[m1+15];
			//R: coadd level 4 check
			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0)|
				(d8>tt0)|(d9>tt0)|(d10>tt0)|(d11>tt0)|(d12>tt0)|(d13>tt0)|(d14>tt0)|(d15>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			d4.xy=d8.xz+d8.yw;d4.zw=d9.xz+d9.yw;d5.xy=d10.xz+d10.yw;d5.zw=d11.xz+d11.yw;
			d6.xy=d12.xz+d12.yw;d6.zw=d13.xz+d13.yw;d7.xy=d14.xz+d14.yw;d7.zw=d15.xz+d15.yw;
			//R: coadd level 5 check
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1)|(d4>tt1)|(d5>tt1)|(d6>tt1)|(d7>tt1);
			//R: next coadd level starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: coadd level 6 check
			was_pulse|=(d0>tt2)|(d1>tt2)|(d2>tt2)|(d3>tt2);
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			if(need_large_FFA_data!=-1){//R: need to save data for large FFA only at beginning of large DM chunk
				if(tid==0){
				//R: Store negative DM sign power data for large FFA
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_neg[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}else if(tid==1){//R: positive DM sign for large FFA
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)]=d0;
					large_pos[need_large_FFA_data+(mchunk<<1)+dchunk*(32768>>10)+1]=d1;
				}
			}

			//R: coadd level 7 check
			was_pulse|=(d0>tt3)|(d1>tt3);
			was_pulse.x|=was_pulse.y|was_pulse.z|was_pulse.w;//result reduction
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			//R: coadd level 8 check
			was_pulse.x|=(d0.x>t.x)|(d0.y>t.x)|(d0.z>t.x)|(d0.w>t.x);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;
			//R: coadd level 9 check,last needed coadd
			was_pulse.x|=(d0.x>t.y)|(d0.y>t.y);
	}
	if(was_pulse.x) results[0]=1;
}
#endif

#else
//R: difference from prev version: smaller registers array (possible more active blocks per compute unit)
//and full use of coadds even if writes will be not float4 (they will be write combined anyway)
//Another possible advantage - smaller memory fetch stride - less or no memory banks/channels conflicts
__kernel void PC_single_pulse_kernel_FFA_update_reduce0(__global const float4* gpu_power,__global float* coadded,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks,
									   const int need_small_FFA_data,//const int need_large_FFA_data,
									   __global float2* small_neg,__global float2* small_pos//,
									   //__global float4* large_neg,__global float4* large_pos
									   )
{
	//R: each work-item will handle separate data_chunk/dm/sign power array through all needed coadds
	//R: to avoid host to device data transfers (very costly for current implementation)
	//R: thresholds and best power values will be maintained separately for CPU and GPU
	//R: thresholds are constant, but best pulses will need update.
	//R: need_*_FFA_data are non-zero (if needed) offsets into gpu_power buffer
	//R: we store small FFA buffers (negative first, then positive) right after coadds buffer
	//R: (coadds buffer take same size as initial power array thought it actually slightly smaller, just to simplify
	//R: buffer size calculations for now) then larger FFA buffers go, first negative one, next positive, to match with
	//R: signs order for GPU dechirping.
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);
//	uint len4 = 32768 >> 3;
//	uint4 coadd_offsets=(uint4)(0);
//	uint4 coadd_begins=(uint4)(0);
	uint next_begin,next_offset;
	float4 t=thresh[0];//thresholds for first 4 coadds
	float4 tt0=(float4)(t.x);//R:will keep current threshold for vector operations
	float4 tt1=(float4)(t.y);//R:will keep current threshold for vector operations
	float4 tt2=(float4)(t.z);//R:will keep current threshold for vector operations
	float4 tt3=(float4)(t.w);//R:will keep current threshold for vector operations
	//uint result=0;
	//int4 pulse0=(int4)(0);//R: result of comparison
	//int4 pulse1=(int4)(0);//R: result of comparison
	//int4 pulse2=(int4)(0);//R: result of comparison
	//int4 pulse3=(int4)(0);//R: result of comparison
	int4 was_pulse=(int4)(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0,d1,d2,d3,d4,d5,d6,d7;
//R: no arrays in registers still (AMD still had to implement it) so store data in separate registers
	//R: Coadded values storage address arithmetics now greatly simplified
	//R: only 1 coadd level stored per kernel call so max 2 coadds need to be addressed
	next_begin=num_dchunks*32*(32768>>3);//R: area just after initial power array
	next_offset=next_begin+(dchunk*32+tid)*(32768>>8);//R: don't account for mchunk offset

	//coadd_begins.x=0;//R:initial array - no need to write, in gpu_power domain
	//coadd_begins.y=num_dchunks*32*(32768>>3);
	//coadd_begins.z=coadd_begins.y+num_dchunks*(32768>>4)*32;
	//coadd_begins.w=coadd_begins.z+num_dchunks*(32768>>5)*32;
	//next_begin=coadd_begins.w+num_dchunks*(32768>>6)*32;

	//coadd_offsets.x=coadd_begins.x+(dchunk*32+tid)*(32768>>3);//l==0
	//coadd_offsets.y=coadd_begins.y+(dchunk*32+tid)*(32768>>4);
	//coadd_offsets.z=coadd_begins.z+(dchunk*32+tid)*(32768>>5);
	//coadd_offsets.w=coadd_begins.w+(dchunk*32+tid)*(32768>>6);
	//next_offset=next_begin+(dchunk*32+tid)*(32768>>7);
	//for(int m=0;m<len4;m+=16)
	{//R:loading data piece of 16*4 samples
		int m1=mchunk<<3;//R:reading corresponding mchunk (8*4 samples here) from power array - coadd level 0
			d0=gpu_power[m1];d1=gpu_power[m1+1];d2=gpu_power[m1+2];d3=gpu_power[m1+3];
			d4=gpu_power[m1+4];d5=gpu_power[m1+5];d6=gpu_power[m1+6];d7=gpu_power[m1+7];

			was_pulse|=(d0>tt0)|(d1>tt0)|(d2>tt0)|(d3>tt0)|(d4>tt0)|(d5>tt0)|(d6>tt0)|(d7>tt0);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			d2.xy=d4.xz+d4.yw;d2.zw=d5.xz+d5.yw;d3.xy=d6.xz+d6.yw;d3.zw=d7.xz+d7.yw;
			//R: next coadd level ready to check for pulses
			was_pulse|=(d0>tt1)|(d1>tt1)|(d2>tt1)|(d3>tt1);
			//R: coadd level 2 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;d1.xy=d2.xz+d2.yw;d1.zw=d3.xz+d3.yw;
			was_pulse|=(d0>tt2)|(d1>tt2);
			//R: coadd level 3 starts
			d0.xy=d0.xz+d0.yw;d0.zw=d1.xz+d1.yw;
			was_pulse|=(d0>tt3);
			d0.xy=d0.xz+d0.yw;
			t=thresh[1];
			was_pulse.xy|=was_pulse.zw;
			was_pulse.xy|=(d0.xy>t.xx);
			//R: level 4 coadd saves data for small FFA sometimes
			if(need_small_FFA_data!=-1){//R: after some data chunk we will not store data for small FFA
				if(tid==0){//R: array of float2 now
				//R: Store negative DM sign power data for small FFA
					small_neg[(need_small_FFA_data<<1)+mchunk+dchunk*(32768>>7)]=d0.xy;
				}else if(tid==1){//R: positive DM sign for small FFA
					small_pos[(need_small_FFA_data<<1)+mchunk+dchunk*(32768>>7)]=d0.xy;
				}
			}
			d0.x+=d0.y;
			was_pulse.x|=was_pulse.y|(d0.x>t.y);
			m1=mchunk+next_offset;
			coadded[m1]=d0.x;

	}
	if(was_pulse.x) results[0]=1;
}
#endif
/*
__kernel void PC_single_pulse_kernel_FFA_update_reduce2(__global float4* gpu_power,
									   __constant float4* thresh, __global uint* results,
									   const int num_dchunks
									   )
{
	uint tid=get_global_id(0);
	uint dchunk=get_global_id(1);
	uint mchunk=get_global_id(2);
//R: now lets repeat for next few coadd levels and refill registers array from GPU memory
//R: only 2 levels remain so to improve parallelization and reduce register pressure let's use smaller registers array
	float4 t=thresh[2];
	//tt2=(float4)(t.z);
	//tt3=(float4)(t.w);
	int was_pulse=(0);////R: simple yes/no flag: was single pulse found somewhere or not;
	//vector for vectorized ops
	float4 d0;
//	pulse0=(int4)(0);
//	pulse1=(int4)(0);
//	pulse2=(int4)(0);
//	pulse3=(int4)(0);
	int initial_offset=num_dchunks*32*(32768>>3)+num_dchunks*(32768>>7)*32+(dchunk*32+tid)*(32768>>11);//R: will fetch data stored at the end of prev coadd bunch loop
	d0=gpu_power[initial_offset];
			//R: coadd level 8 check
			was_pulse=(d0.x>t.x)|(d0.y>t.x)|(d0.z>t.x)|(d0.w>t.x);
			//R: now do coadd and save coadded piece in registers
			d0.xy=d0.xz+d0.yw;
			//R: coadd level 9 check,last needed coadd
			was_pulse|=(d0.x>t.y)|(d0.y>t.y);
	if(was_pulse) results[0]=1;
}
*/
//R: block of support functions for df64 DP emulation
float __fmul_rn(float a, float b){return a*b;}
float __fadd_rn(float a, float b){return a+b;}
float __float2int_rn(float a){return rint(a);}
float __int2float_rn(int a){return (float)a;}
float2 split(float a)
{
	const float split= 4097;
	float t = __fmul_rn(a,split);
	float2 r;
	r.x = t-(t-a);
	r.y = a - r.x;
	return r;
}
 float2 twoProd(float a, float b)
{
	float2 r;
	r.x = __fmul_rn(a,b);
	float2 aS = split(a);
	float2 bS = split(b);
	r.y = ( ( __fmul_rn(aS.x,bS.x) - r.x )
				+ __fmul_rn(aS.x,bS.y) + __fmul_rn(aS.y,bS.x) )
				+ __fmul_rn(aS.y,bS.y);
	return r;
}

float2 quickTwoSum(float a, float b)
{
	float2 r;
	r.x = a + b;
	r.y = b - (r.x-a);
	return r;
}

float2 df64_mult(float2 a, float2 b)
{
	float ah=a.x,al=a.y,bh=b.x,bl=b.y;
	float2 p; 
	p = twoProd(ah,bh);
	p.y += ah*bl;
	p.y += al*bh;
	p = quickTwoSum(p.x,p.y);
	return p;
}

#if 0
/*R: disadvantage of this kernel:
same offset calculations are performed for each element in output array, that's redundant.
Maybe merging offset calculation will be viable with 1D kernel variant.
*/
__kernel void GPU_fetch_array_kernel_df64_cl(__global float* src, __constant float2* periods, __global int* f_int, __global float4* dest,
		const uint offset, const uint stride){
	uint threadID=get_global_id(0)+offset;
	uint j=get_global_id(1);
	int k=0;
	int l=0;
	float4 acc=(float4)0.f;
	int n_per=f_int[threadID];
	float2 period=periods[threadID];
	for(k=0;k<n_per;k++){
		float2 jj=split(__int2float_rn(k));
		float2 tmp=df64_mult(period,jj);
		float int_part=trunc(tmp.x);
		tmp.x-=int_part;
		tmp.x+=tmp.y;
			//l=offsets[k*stride+threadID];
			l=((int)int_part)+((int)rint(tmp.x));
			l+=(4*j);//R: index to data array computed
			acc.x+=src[l];
			acc.y+=src[l+1];
			acc.z+=src[l+2];
		 	acc.w+=src[l+3];
#if 0
	if(l >= 65534){
		printf("threadID=%d; j=%d; n_per=%d\n",threadID,j,n_per);
	 	printf("l=%d; k=%d; array value:(%.4g, %.4g, %.4g, %.4g)\n",l,k,src[l],src[l+1],src[l+2],src[l+3]);
	}
#endif
	}
	dest[threadID*480+j]=acc;
}
#elif 1
#if TWIN_FFA
/*R: some statistics for FFA used in v6 AP:
data buffer: 4*64k small, 4*256k large
fetched array: 4*(120*4)=1920b per thread small; 4*(479*4)=7664b per thread large
offsets: 273 per thread 
*/
//R: these 3 defines set vector size of fetch kernel
typedef float4 ftype; 
#define FETCH vload4
#define FSHIFT 2 // 0 - scalar; 2 - float4; 3 - float 8; 4 - float16
#define FD (480) //(480<<2) - scalar/float; (480) - float4; (480>>1) - float8; (480>>2) - float16
#if OCL_X4_FETCH
__kernel void GPU_fetch_array_kernel_twin_x4_cl(__global float* src, __global float* src_twin, 
											  __global int4* offsets,  __global int4* f_int, 
											 __global ftype* dest, __global ftype* dest_twin,
		const uint offset, const uint stride){
	int threadID=get_global_id(0)+offset;
	int j=get_global_id(1);
	if(j>=FD)return; //R:matrix bounds guard
	int4 l=(int4)0;
	ftype acc_x=(ftype)0.f,acc_twin_x=(ftype)0.f;
	ftype acc_y=(ftype)0.f,acc_twin_y=(ftype)0.f;
	ftype acc_z=(ftype)0.f,acc_twin_z=(ftype)0.f;
	ftype acc_w=(ftype)0.f,acc_twin_w=(ftype)0.f;
	int4 n_per=f_int[threadID];
//R: will use ordering of freqs. Ordering results in increase of n_per (and decrease of array size).
//hence always n_per.x<n_per.y<...
	for(size_t k=0;k<n_per.x;k++){
			l=offsets[k*stride+threadID];
			l+=(int4)(j<<FSHIFT);//R: index to data array computed
			__global float* s;
			s=src+l.x;
			acc_x+=FETCH(0,s);
			s=src_twin+l.x;
			acc_twin_x+=FETCH(0,s);
			s=src+l.y;
			acc_y+=FETCH(0,s);
			s=src_twin+l.y;
			acc_twin_y+=FETCH(0,s);
			s=src+l.z;
			acc_z+=FETCH(0,s);
			s=src_twin+l.z;
			acc_twin_z+=FETCH(0,s);
			s=src+l.w;
			acc_w+=FETCH(0,s);
			s=src_twin+l.w;
			acc_twin_w+=FETCH(0,s);
	}
	for(size_t k=n_per.x;k<n_per.y;k++){
			l=offsets[k*stride+threadID];
			l+=(int4)(j<<FSHIFT);//R: index to data array computed
			__global float* s;
			s=src+l.y;
			acc_y+=FETCH(0,s);
			s=src_twin+l.y;
			acc_twin_y+=FETCH(0,s);
			s=src+l.z;
			acc_z+=FETCH(0,s);
			s=src_twin+l.z;
			acc_twin_z+=FETCH(0,s);
			s=src+l.w;
			acc_w+=FETCH(0,s);
			s=src_twin+l.w;
			acc_twin_w+=FETCH(0,s);
	}
	for(size_t k=n_per.y;k<n_per.z;k++){
			l=offsets[k*stride+threadID];
			l+=(int4)(j<<FSHIFT);//R: index to data array computed
			__global float* s;
			s=src+l.z;
			acc_z+=FETCH(0,s);
			s=src_twin+l.z;
			acc_twin_z+=FETCH(0,s);
			s=src+l.w;
			acc_w+=FETCH(0,s);
			s=src_twin+l.w;
			acc_twin_w+=FETCH(0,s);
	}
	for(size_t k=n_per.z;k<n_per.w;k++){
			l=offsets[k*stride+threadID];
			l+=(int4)(j<<FSHIFT);//R: index to data array computed
			__global float* s;
			s=src+l.w;
			acc_w+=FETCH(0,s);
			s=src_twin+l.w;
			acc_twin_w+=FETCH(0,s);
	}
	threadID<<=2;
	threadID*=FD;
	int4 index;
	index.x=threadID+(j);
	index.y=threadID+FD+(j);
	index.z=threadID+FD+FD+(j);
	index.w=threadID+FD+FD+FD+(j);
	dest[index.x]=acc_x;
	dest_twin[index.x]=acc_twin_x;
	dest[index.y]=acc_y;
	dest_twin[index.y]=acc_twin_y;
	dest[index.z]=acc_z;
	dest_twin[index.z]=acc_twin_z;
	dest[index.w]=acc_w;
	dest_twin[index.w]=acc_twin_w;
}
#elif OCL_1D_FETCH
#define FARRAY_SIZE 12
#define CACHE_SIZE 4096
void fetch_caches(__global float* src, __global float* src_twin, __local float* src_cache, __local float* src_twin_cache){
	size_t tid=get_local_id(0);
	vstore4(vload4(tid,src),tid,src_cache);
	vstore4(vload4(tid,src_twin),tid,src_twin_cache);
	tid+=256;
	vstore4(vload4(tid,src),tid,src_cache);
	vstore4(vload4(tid,src_twin),tid,src_twin_cache);
	tid+=256;
	vstore4(vload4(tid,src),tid,src_cache);
	vstore4(vload4(tid,src_twin),tid,src_twin_cache);
	tid+=256;
	vstore4(vload4(tid,src),tid,src_cache);
	vstore4(vload4(tid,src_twin),tid,src_twin_cache);
	barrier(CLK_LOCAL_MEM_FENCE);
	//event_t ev[2];
	//ev[0]=async_work_group_copy (src_cache,src,4096,ev[0]);
	//ev[1]=async_work_group_copy (src_twin_cache,src_twin,4096,ev[1]);
	//wait_group_events(2,ev);
}
/*R: some discussion of optimal configuration for fetch kernel:
Accordingly to http://devgurus.amd.com/message/1303698#1303698 AMD GPU not able to swap waves belonging to the same
workgroup. This means that with only 4 waves per CU hiding memory latency by waves swap is impossible.
Hence minimum 8 (and not 4) waves in fly (2 workgroups) per CU should be for waves swapping occur.
*/
/*
__kernel __attribute__((reqd_work_group_size(256, 1, 1))) 
void GPU_fetch_array_kernel_twin_1D_cl(__global float* src, __global float* src_twin, 
											  __global int* offsets,  __global int* f_int, 
											 __global ftype* dest, __global ftype* dest_twin,
		const uint offset, const uint stride, const uint max_per_int){
	size_t threadID=get_global_id(0)+offset;
//	size_t j=get_global_id(1);
//	if(j>=fd)return; //R:matrix bounds guard
	size_t l;
	ftype acc[FARRAY_SIZE],	acc_twin[FARRAY_SIZE];
	__local float src_cache[CACHE_SIZE];
	__local float src_twin_cache[CACHE_SIZE];
	fetch_caches(src,src_twin,src_cache,src_twin_cache);
	//event_t ev[2];
	//ev[0]=async_work_group_copy (src_cache,src,4096,ev[0]);
	//ev[1]=async_work_group_copy (src_twin_cache,src_twin,4096,ev[1]);
	size_t n_per=f_int[threadID];
	for(size_t outer_loop=0;outer_loop<max_per_int;outer_loop+=FARRAY_SIZE){
		for(size_t j=0;j<FARRAY_SIZE;j++){acc[j]=0.f;acc_twin[j]=0.f;}
	//wait_group_events(2,ev);
    for(size_t k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			for(size_t j=0;j<FARRAY_SIZE;j++){
				size_t displace=l+((outer_loop+j)<<FSHIFT);
				if(displace<=(CACHE_SIZE-(1<<FSHIFT))){
					__local float* s=src_cache+displace;
					acc[j]+=FETCH(0,s);
					s=src_twin_cache+displace;
					acc_twin[j]+=FETCH(0,s);
				}else{
					__global float* s=src+displace;
					acc[j]+=FETCH(0,s);
					s=src_twin+displace;
					acc_twin[j]+=FETCH(0,s);
				}
	 		}
	 }
	 for(size_t j=0;j<FARRAY_SIZE;j++){
	 	size_t index=threadID*FD+(outer_loop+j);
	 	dest[index]=acc[j];
	 	dest_twin[index]=acc_twin[j];
	 }
	}

}*/
__kernel __attribute__((reqd_work_group_size(64, 1, 1))) 
void GPU_fetch_array_kernel_twin_1D_persistent_cl(__global float* src, __global float* src_twin, 
											  __global int* offsets,  __global int* f_int, 
											 __global ftype* dest, __global ftype* dest_twin,
		const uint offset, const uint stride, const uint max_per_int,const uint max_freqs_num){
	size_t threadID=get_global_id(0)+offset;
//	size_t j=get_global_id(1);
//	if(j>=fd)return; //R:matrix bounds guard
	size_t l;
	ftype acc[FARRAY_SIZE],	acc_twin[FARRAY_SIZE];
	__local float src_cache[CACHE_SIZE];
	__local float src_twin_cache[CACHE_SIZE];
	fetch_caches(src,src_twin,src_cache,src_twin_cache);
	//event_t ev[2];
	//ev[0]=async_work_group_copy (src_cache,src,4096,ev[0]);
	//ev[1]=async_work_group_copy (src_twin_cache,src_twin,4096,ev[1]);
	while(threadID<max_freqs_num){
	size_t n_per=f_int[threadID];
	for(size_t outer_loop=0;outer_loop<max_per_int;outer_loop+=FARRAY_SIZE){
		for(size_t j=0;j<FARRAY_SIZE;j++){acc[j]=0.f;acc_twin[j]=0.f;}
	//wait_group_events(2,ev);
    for(size_t k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			for(size_t j=0;j<FARRAY_SIZE;j++){
				size_t displace=l+((outer_loop+j)<<FSHIFT);
				if(displace<=(CACHE_SIZE-(1<<FSHIFT))){
					__local float* s=src_cache+displace;
					acc[j]+=FETCH(0,s);
					s=src_twin_cache+displace;
					acc_twin[j]+=FETCH(0,s);
				}else{
					__global float* s=src+displace;
					acc[j]+=FETCH(0,s);
					s=src_twin+displace;
					acc_twin[j]+=FETCH(0,s);
				}
	 		}
	 }
	 for(size_t j=0;j<FARRAY_SIZE;j++){
	 	size_t index=threadID*FD+(outer_loop+j);
	 	dest[index]=acc[j];
	 	dest_twin[index]=acc_twin[j];
	 }
	}
	threadID+=get_global_size(0);
	}
}
#else
__kernel void GPU_fetch_array_kernel_twin_cl(__global float* src, __global float* src_twin, 
											  __global int* offsets,  __global int* f_int, 
											 __global ftype* dest, __global ftype* dest_twin,
		const uint offset, const uint stride){
	size_t threadID=get_global_id(0)+offset;
	size_t j=get_global_id(1);
	if(j>=FD)return; //R:matrix bounds guard
	size_t l=0;
	ftype acc=(ftype)0.f,acc_twin=(ftype)0.f;
	size_t n_per=f_int[threadID];
	for(size_t k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(j<<FSHIFT);//R: index to data array computed
			__global float* s=src+l;
			acc+=FETCH(0,s);
			s=src_twin+l;
			acc_twin+=FETCH(0,s);
	}
	size_t index=threadID*FD+(j);
	dest[index]=acc;
	dest_twin[index]=acc_twin;
}
#endif

#if 0

#endif //1D fetch kernel

/*
#if 0 //scalar fetch
__kernel void GPU_fetch_array_kernel_twin_cl(__global float* src, __global float* src_twin, 
											  __global int* offsets,  __global int* f_int, 
											 __global float* dest, __global float* dest_twin,
		const uint offset, const uint stride){
	size_t threadID=get_global_id(0)+offset;
	size_t j=get_global_id(1);
	size_t l=0;
	float acc=(float)0.f,acc_twin=(float)0.f;
	//float16 acc1=(float16)0.f,acc_twin1=(float16)0.f;
	size_t n_per=f_int[threadID];
	for(size_t k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(j);//R: index to data array computed
			//__global float* s=src+l;
			acc+=src[l];//vload4(0,s);
			//acc1+=vload16(1,s);
			//s=src_twin[l];
			acc_twin+=src_twin[l];//vload4(0,s);
			//acc_twin1+=vload16(1,s);
	}
	size_t index=threadID*(480<<2)+(j);
	dest[index]=acc;
	dest_twin[index]=acc_twin;
	//index++;
	//dest[index]=acc1;
	//dest_twin[index]=acc_twin1;
}
#elif 0 //vload4 fetch
__kernel void GPU_fetch_array_kernel_twin_cl(__global float* src, __global float* src_twin, 
											  __global int* offsets,  __global int* f_int, 
											 __global float4* dest, __global float4* dest_twin,
		const uint offset, const uint stride){
	size_t threadID=get_global_id(0)+offset;
	size_t j=get_global_id(1);
	size_t l=0;
	float4 acc=(float4)0.f,acc_twin=(float4)0.f;
	size_t n_per=f_int[threadID];
	for(size_t k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(j<<2);//R: index to data array computed
			__global float* s=src+l;
			acc+=vload4(0,s);
			s=src_twin+l;
			acc_twin+=vload4(0,s);
	}
	size_t index=threadID*480+(j);
	dest[index]=acc;
	dest_twin[index]=acc_twin;
}
#elif 0 //vload8 fetch
__kernel void GPU_fetch_array_kernel_twin_cl(__global float* src, __global float* src_twin, 
											  __global int* offsets,  __global int* f_int, 
											 __global float8* dest, __global float8* dest_twin,
		const uint offset, const uint stride){
	size_t threadID=get_global_id(0)+offset;
	size_t j=get_global_id(1);
	size_t l=0;
	float8 acc=(float8)0.f,acc_twin=(float8)0.f;
	size_t n_per=f_int[threadID];
	for(size_t k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(j<<3);//R: index to data array computed
			__global float* s=src+l;
			acc+=vload8(0,s);
			s=src_twin+l;
			acc_twin+=vload8(0,s);
	}
	size_t index=threadID*(480>>1)+(j);
	dest[index]=acc;
	dest_twin[index]=acc_twin;
}
#else //vload16 fetch ("vload32" via 2 vload16 was quite slower in all tests so far)
__kernel void GPU_fetch_array_kernel_twin_cl(__global float* src, __global float* src_twin, 
											  __global int* offsets,  __global int* f_int, 
											 __global float16* dest, __global float16* dest_twin,
		const uint offset, const uint stride){
	size_t threadID=get_global_id(0)+offset;
	size_t j=get_global_id(1);
	size_t l=0;
	float16 acc=(float16)0.f,acc_twin=(float16)0.f;
	size_t n_per=f_int[threadID];
	for(size_t k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(j<<4);//R: index to data array computed
			__global float* s=src+l;
			acc+=vload16(0,s);
			//acc1+=vload16(1,s);
			s=src_twin+l;
			acc_twin+=vload16(0,s);
			//acc_twin1+=vload16(1,s);
	}
	size_t index=threadID*(480>>2)+(j);
	dest[index]=acc;
	dest_twin[index]=acc_twin;
	//index++;
	//dest[index]=acc1;
	//dest_twin[index]=acc_twin1;
}
#endif
*/
#else
__kernel void GPU_fetch_array_kernel_cl(__global float* src, __global int* offsets, __global int* f_int, __global float4* dest,
		const uint offset, const uint stride){
	uint threadID=get_global_id(0)+offset;
	uint j=get_global_id(1);
	int k=0;
	int l=0;
	float4 acc=(float4)0.f;
	int n_per=f_int[threadID];
	for(k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(4*j);//R: index to data array computed
			acc.x+=src[l];
			acc.y+=src[l+1];
			acc.z+=src[l+2];
		 	acc.w+=src[l+3];
#if 0
	if(l >= 65534){
		printf("threadID=%d; j=%d; n_per=%d\n",threadID,j,n_per);
	 	printf("l=%d; k=%d; array value:(%.4g, %.4g, %.4g, %.4g)\n",l,k,src[l],src[l+1],src[l+2],src[l+3]);
	}
#endif

	}
	dest[threadID*480+j]=acc;
}
#endif
#elif 0
/*R: this kernel uses non-vectorised memory accesses in preparation to 1D kernel variant
Slower (~54.0ms vs ~18.7ms) than vectorized one on C-60 APU
*/
__kernel void GPU_fetch_array_kernel_fl_cl(__global float* src, __global int* offsets, __global int* f_int, __global float* dest,
		const uint offset, const uint stride){
	uint threadID=get_global_id(0)+offset;
	uint j=get_global_id(1);
	int k=0;
	int l=0;
	float acc=0.f;
	int n_per=f_int[threadID];
	for(k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(j);//R: index to data array computed
			acc+=src[l];
#if 0
	if(l >= 65534){
		printf("threadID=%d; j=%d; n_per=%d\n",threadID,j,n_per);
	 	printf("l=%d; k=%d; array value:(%.4g, %.4g, %.4g, %.4g)\n",l,k,src[l],src[l+1],src[l+2],src[l+3]);
	}
#endif

	}
	dest[threadID*(480<<2)+j]=acc;
}
#elif 0
/*R: few workitems of original kernel merged into one. This allows to reduce number of reads from offsets matrix*/
#define FETCH_UNROLL_FACTOR 10
__kernel void GPU_fetch_array_kernel_cl(__global float* src, __global int* offsets, __global int* f_int, __global float4* dest,
		const uint offset, const uint stride){
	uint threadID=get_global_id(0)+offset;
	uint j=get_global_id(1)*FETCH_UNROLL_FACTOR;
	int k=0;
	int l=0;
	float4 acc=(float4)0.f,acc1=(float4)0.f,acc2=(float4)0.f,acc3=(float4)0.f,
		acc4=(float4)0.f,acc5=(float4)0.f,acc6=(float4)0.f,acc7=(float4)0.f,acc8=(float4)0.f,acc9=(float4)0.f;
	
	//float acc[FETCH_UNROLL_FACTOR];
	//for(int i=0;i<FETCH_UNROLL_FACTOR;i++) acc[i]=0;
	int n_per=f_int[threadID];
	for(k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			l+=(j<<2);//R: index to data array computed
			/*for (int i=0;i<FETCH_UNROLL_FACTOR;i++)
				acc[i]+=src[l+i];*/
			acc.x+=src[l];
			acc.y+=src[l+1];
			acc.z+=src[l+2];
		 	acc.w+=src[l+3];
			acc1.x+=src[l+4];
			acc1.y+=src[l+5];
			acc1.z+=src[l+6];
		 	acc1.w+=src[l+7];
			acc2.x+=src[l+8];
			acc2.y+=src[l+9];
			acc2.z+=src[l+10];
		 	acc2.w+=src[l+11];
			acc3.x+=src[l+12];
			acc3.y+=src[l+13];
			acc3.z+=src[l+14];
		 	acc3.w+=src[l+15];
			acc4.x+=src[l+16];
			acc4.y+=src[l+17];
			acc4.z+=src[l+18];
		 	acc4.w+=src[l+19];
			acc5.x+=src[l+20];
			acc5.y+=src[l+21];
			acc5.z+=src[l+22];
		 	acc5.w+=src[l+23];
			acc6.x+=src[l+24];
			acc6.y+=src[l+25];
			acc6.z+=src[l+26];
		 	acc6.w+=src[l+27];
			acc7.x+=src[l+28];
			acc7.y+=src[l+29];
			acc7.z+=src[l+30];
		 	acc7.w+=src[l+31];
			acc8.x+=src[l+32];
			acc8.y+=src[l+33];
			acc8.z+=src[l+34];
		 	acc8.w+=src[l+35];
			acc9.x+=src[l+36];
			acc9.y+=src[l+37];
			acc9.z+=src[l+38];
		 	acc9.w+=src[l+39];
#if 0
	if(l >= 65534){
		printf("threadID=%d; j=%d; n_per=%d\n",threadID,j,n_per);
	 	printf("l=%d; k=%d; array value:(%.4g, %.4g, %.4g, %.4g)\n",l,k,src[l],src[l+1],src[l+2],src[l+3]);
	}
#endif

	}
	//for(int i=0;i<FETCH_UNROLL_FACTOR;i++) dest[threadID*(480*4)+j+i]=acc[i];
	dest[threadID*480+j]=acc;
	dest[threadID*480+j+1]=acc1;
	dest[threadID*480+j+2]=acc2;
	dest[threadID*480+j+3]=acc3;
	dest[threadID*480+j+4]=acc4;
	dest[threadID*480+j+5]=acc5;
	dest[threadID*480+j+6]=acc6;
	dest[threadID*480+j+7]=acc7;	
	dest[threadID*480+j+8]=acc8;
	dest[threadID*480+j+9]=acc9;
}
#elif 0
/*R:  1D fetch kernel variant with uniform array lengths
*/
__kernel void GPU_fetch_array_kernel_fl_cl(__global float* src, __global int* offsets, __global int* f_int, __global float* dest,
		const uint offset, const uint stride, const int length){
	uint threadID=get_global_id(0)+offset;
	uint j=get_global_id(1);
	int k=0;
	int l=0;
	float acc=0.f;
	int n_per=f_int[threadID];
	for(k=0;k<n_per;k++){
			l=offsets[k*stride+threadID];
			for(int j=0;j<length;j++)
			//l+=(j);//R: index to data array computed
				acc+=src[l+j];

	}
	dest[threadID*480+j]=acc;
}

#endif

#if TWIN_FFA
__kernel void GPU_compare_with_threshold_kernel_twin_cl(const int size, __global float4* input,__global float4* input_twin, 
														__global float* thresholds, const int level, 
														__global int* output,__global int* output_twin, const uint stride)
{
	int threadID=get_global_id(0);
	float threshold=thresholds[level+threadID*12];
	float4 p,p_twin;
	int i=size-1;
	//int was_signal=0;
	int bin=99999,bin_twin=99999;
	//int ln=(len[threadID]+3)>>2;
	for(;i>=0;i--){
		p=input[i+threadID*stride];
		p_twin=input_twin[i+threadID*stride];
		if(p.w>=threshold){
			//was_signal++;
			bin=4*i+3;
		}
		if(p.z>=threshold){
			//was_signal++;
			bin=4*i+2;
		}
		if(p.y>=threshold){
			//was_signal++;
			bin=4*i+1;
		}
		if(p.x>=threshold){
			//was_signal++;
			bin=4*i;
		}
		if(p_twin.w>=threshold){
			//was_signal++;
			bin_twin=4*i+3;
		}
		if(p_twin.z>=threshold){
			//was_signal++;
			bin_twin=4*i+2;
		}
		if(p_twin.y>=threshold){
			//was_signal++;
			bin_twin=4*i+1;
		}
		if(p_twin.x>=threshold){
			//was_signal++;
			bin_twin=4*i;
		}
	}
	//output.x=was_signal;
	output[threadID]=bin;
	output_twin[threadID]=bin_twin;
}
#else
__kernel void GPU_compare_with_threshold_kernel_cl(const int size, __global float4* input, __global float* thresholds,
												   const int level, __global int* output, const uint stride)
{
	int threadID=get_global_id(0);
	float threshold=thresholds[level+threadID*12];
	float4 p;
	int i=size-1;
	//int was_signal=0;
	int bin=99999;
	//int ln=(len[threadID]+3)>>2;
	for(;i>=0;i--){
		p=input[i+threadID*stride];
		if(p.w>=threshold){
			//was_signal++;
			bin=4*i+3;
		}
		if(p.z>=threshold){
			//was_signal++;
			bin=4*i+2;
		}
		if(p.y>=threshold){
			//was_signal++;
			bin=4*i+1;
		}
		if(p.x>=threshold){
			//was_signal++;
			bin=4*i;
		}
	}
	//output.x=was_signal;
	output[threadID]=bin;
}
#endif
//R: TODO: currenly returned threadID can be used only as some debug info (order in what threads will update single memory location is undefined
//so, 2 possibilities for optimization here:
//1: to make search kernel fully agnostic vregarding pulse location, then 2D kernel with each workitem processing single element can be used
//(threshold better to cache into local memory then)
//2: to reserve separate location for each thread (2D kernel still possible). Then  will be possible to narrow re-processing area in case of found pulse
#if TWIN_FFA
__kernel void GPU_PC_compare_with_threshold_kernel_twin_cl(const int size, __global float4* input,__global float4* input_twin, 
														   __global float* thresholds, const int level,
														   __global int* output, 
														   const uint stride,__global int* array_size)
{
	int threadID=get_global_id(0);
	float threshold=thresholds[level+threadID*12];
	float4 p,p_twin;
	int i=size-1;//TODO: real particular array size can be used instead of generic maximum possible size
	//int was_signal=0;
	int bin=99999;
	//int ln=(len[threadID]+3)>>2;
	for(;i>=0;i--){
		p=input[i+threadID*stride];	
		p_twin=input_twin[i+threadID*stride];
		if(p.w>=threshold || p_twin.w>=threshold){
			bin=4*i+3;
		}
		if(p.z>=threshold || p_twin.z>=threshold){
			bin=4*i+2;
		}
		if(p.y>=threshold || p_twin.y>=threshold){
			bin=4*i+1;
		}
		if(p.x>=threshold || p_twin.x>=threshold){
			bin=4*i;
		}
	}
	//R: actually, with twin mod both buffers will be re-processed anyway so no matter in wich buffer signal was found
	if(bin<array_size[threadID]){//TODO: if real array size used then just compare that bin had changed.
		output[0]=1+threadID;
	}
}
#else
__kernel void GPU_PC_compare_with_threshold_kernel_cl(const int size, __global float4* input, __global float* thresholds,
												   const int level, __global int* output, const uint stride,__global int* array_size)
{
	int threadID=get_global_id(0);
	float threshold=thresholds[level+threadID*12];
	float4 p;
	int i=size-1;//TODO: real particular array size can be used instead of generic maximum possible size
	//int was_signal=0;
	int bin=99999;
	//int ln=(len[threadID]+3)>>2;
	for(;i>=0;i--){
		p=input[i+threadID*stride];
		if(p.w>=threshold){
			//was_signal++;
			bin=4*i+3;
		}
		if(p.z>=threshold){
			//was_signal++;
			bin=4*i+2;
		}
		if(p.y>=threshold){
			//was_signal++;
			bin=4*i+1;
		}
		if(p.x>=threshold){
			//was_signal++;
			bin=4*i;
		}
	}
	//output.x=was_signal;
	if(bin<array_size[threadID]){//TODO: if real array size used then just compare that bin had changed.
		output[0]=1+threadID;
	}
}
#endif
#if 1
#if TWIN_FFA
__kernel void GPU_coadd_kernel_twin_cl(__global float4* src,__global float4* src_twin, 
									   __global float4* dest,__global float4* dest_twin, 
									   const uint stride)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	uint new_stride;
	float4 o,o_twin;
	float4 i2,i2_twin;
	float4 i21,i21_twin;
		i2=src[2*i+threadID*stride];
		i21=src[2*i+1+threadID*stride];
		i2_twin=src_twin[2*i+threadID*stride];
		i21_twin=src_twin[2*i+1+threadID*stride];
		o.xy=i2.xz+i2.yw;
		o.zw=i21.xz+i21.yw;
		o_twin.xy=i2_twin.xz+i2_twin.yw;
		o_twin.zw=i21_twin.xz+i21_twin.yw;
		if(stride>120){
			new_stride=(stride>>1);
		}else{
			new_stride=stride;
		}
		dest[i+threadID*(new_stride)]=o;		
		dest_twin[i+threadID*(new_stride)]=o_twin;
}
#else
__kernel void GPU_coadd_kernel_cl(__global float4* src, __global float4* dest, const uint stride)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	uint new_stride;
	float4 o;
	float4 i2;
	float4 i21;
		i2=src[2*i+threadID*stride];
		i21=src[2*i+1+threadID*stride];
		o.xy=i2.xz+i2.yw;
		o.zw=i21.xz+i21.yw;
		if(stride>120){
			new_stride=(stride>>1);
		}else{
			new_stride=stride;
		}
#if 0
		if(threadID==0 && i==0) printf("stride=%d; new_stride=%d\n",stride,new_stride);
#endif
		dest[i+threadID*(new_stride)]=o;
}
#endif
#else
//uje: 1D, if 2D operation does not work correct e.g. on linux with HD4670
__kernel void GPU_coadd_1D_kernel_cl(const uint size, __global float4* src, __global float4* dest, const uint stride) {
  int threadID = get_global_id(0);
  int i = 0;
  uint new_stride;
  float4 o;
  float4 i2;
  float4 i21;
  for(;
      i<4*size;
      i++) {
    i2 = src[2*i+threadID*stride];
    i21 = src[2*i+1+threadID*stride];
    o.xy = i2.xz+i2.yw;
    o.zw = i21.xz+i21.yw;
    if(stride>120) {
      new_stride = (stride>>1);
    } else {
      new_stride = stride;
    }
    dest[i+threadID*new_stride] = o;
  }
}
#endif
#if 0 //USE_OPENCL_NV && !_GNU_SOURCE //R: non-vectorized version
__kernel void GPU_coadd_with_stride_kernel_cl(__global float* src,__global int* stride,__global float* dest,const uint stride1)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	float o=src[i+threadID*(stride1*4)];
	int st=stride[threadID];
	o+=src[i+st+4*stride1*threadID];
	dest[threadID*240*4+i]=o;
}
#else
#if TWIN_FFA
__kernel void GPU_coadd_with_stride_kernel_twin_cl(__global float4* src,__global float4* src_twin,
											  __global int* stride,
											  __global float4* dest,__global float4* dest_twin,
											  const uint stride1)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	float4 o=src[i+threadID*stride1],o_twin=src_twin[i+threadID*stride1];
	int st=stride[threadID];
	int offset=st>>2;
	int s=st-4*(offset);
	float4 i1=src[i+offset+stride1*threadID],i1_twin=src_twin[i+offset+stride1*threadID];
	float4 i2=src[i+offset+1+stride1*threadID],i2_twin=src_twin[i+offset+1+stride1*threadID];
	//if(i >= per_int[treadID]) return;
	if(s==0){
		o+=i1;o_twin+=i1_twin;
	}else if(s==1){
		o.xyz+=i1.yzw;
		o.w+=i2.x;
		o_twin.xyz+=i1_twin.yzw;
		o_twin.w+=i2_twin.x;
	}else if(s==2){
		o.xy+=i1.zw;
		o.zw+=i2.xy;
		o_twin.xy+=i1_twin.zw;
		o_twin.zw+=i2_twin.xy;
	}else if(s==3){
		o.x+=i1.w;
		o.yzw+=i2.xyz;
		o_twin.x+=i1_twin.w;
		o_twin.yzw+=i2_twin.xyz;
	}
	dest[threadID*240+i]=o;
	dest_twin[threadID*240+i]=o_twin;
}
#else
__kernel void GPU_coadd_with_stride_kernel_cl(__global float4* src,__global int* stride,__global float4* dest,const uint stride1)
{
	int threadID=get_global_id(0);
	int i=get_global_id(1);
	float4 o=src[i+threadID*stride1];
	int st=stride[threadID];
	int offset=st>>2;
	int s=st-4*(offset);
	float4 i1=src[i+offset+stride1*threadID];
	float4 i2=src[i+offset+1+stride1*threadID];
	//if(i >= per_int[treadID]) return;
	if(s==0){
		o+=i1;
	}else if(s==1){
		o.xyz+=i1.yzw;
		o.w+=i2.x;
	}else if(s==2){
		o.xy+=i1.zw;
		o.zw+=i2.xy;
	}else if(s==3){
		o.x+=i1.w;
		o.yzw+=i2.xyz;
	}
	dest[threadID*240+i]=o;
}
#endif
#endif
#if 0
//uje: not using float4 with this 1D kernel
__kernel void GPU_coadd_with_stride_1D_kernel_cl(const uint size,__global float* src,__global int* stride,__global float* dest,const uint stride1) {
  int threadID=get_global_id(0);
  int i=0;
  float o;
  int st=stride[threadID];
  for(;
      i<size;
      i++) {
    o=src[i+4*stride1*threadID];
    o+=src[i+st+4*stride1*threadID];
    dest[threadID*240*4+i]=o;
  }
}
#endif
__kernel void GPU_change_array_sizes_kernel_cl(__global int* array_sizes){
	array_sizes[get_global_id(0)] >>=1;
}
__kernel void GPU_change_array_sizes_per_int_kernel_cl(__global int* array_sizes,__global int* overlap){
	const uint tid=get_global_id(0);
	int v=array_sizes[tid];
	overlap[tid] =  (v >> 1);
    array_sizes[tid] = ((v + 1)  >> 1);
}
#if defined(USE_OPENCL_NV) && defined(OCL_FERMI) 
#pragma OPENCL EXTENSION cl_khr_fp64 : enable  
__kernel void PopulateFetchOffsets_kernel_DP_cl(__global double* periods,__global int* offsets){
	int tid=get_global_id(0);
	int j=get_global_id(1);
	offsets[tid+j*get_global_size(0)]=(int)(periods[tid]*j + 0.5);
}
#endif
#if USE_OCL_DP //R: take advantage of DP card if host app supports this path
#if defined(cl_khr_fp64)   //R:double precision chirp here
#pragma OPENCL EXTENSION cl_khr_fp64 : enable  
#endif
#if defined(cl_amd_fp64)
#pragma OPENCL EXTENSION cl_amd_fp64 : enable  
#endif
//R: f_int can be in constant space if threads number will be less than 16k always 
__kernel void PopulateTresholdTable_kernel9t(int dm_chunk, __global int* f_int,
										 double thresh_x_intercept,
										 double thresh_x_slope,__global float* thresh_tbl){
	int threadID=get_global_id(0);
	int ttidx_pre =gert_global_id(1);
	//int nfb= ((int)freq[threadID]);  
	int nfb=f_int[threadID];
	int ncb= (nfb*dm_chunk)<<ttidx_pre;
	double thresh_x;
	nfb=nfb<<ttidx_pre;
          thresh_x = thresh_x_intercept + thresh_x_slope * (double)nfb;
          thresh_tbl = (float)(ncb + (int)(thresh_x *(double) sqrt(2.0f * (float) ncb)));
}
#else //R: FP-only implementation if possible
__kernel void PopulateTresholdTable_kernel9t_df64_cl(int dm_chunk, __global int* f_int,
										 float2 thresh_x_intercept,//R:in split format
										 float2 thresh_x_slope,//R: in split format
										 __global float* thresh_tbl){
	int threadID=get_global_id(0);
	int ttidx_pre =get_global_id(1);
	//int nfb= ((int)freq[threadID]);  
	int nfb=f_int[threadID];
	int ncb= (nfb*dm_chunk)<<ttidx_pre;
	float2 thresh_x;//R:in split format
	nfb=nfb<<ttidx_pre;
	int tmp=ncb;
    thresh_x = thresh_x_intercept + df64_mult(thresh_x_slope, split(__int2float_rn(nfb)));
	float2 tmp_df64=df64_mult(thresh_x,split(sqrt(2.0f * (float) ncb)));
	tmp+=(int)(tmp_df64.x+tmp_df64.y);
	thresh_tbl[threadID*get_global_size(1)+ttidx_pre] = (float)tmp;
}
__kernel void PopulateFetchOffsets_kernel_df64_cl(__global float2* periods,__global int* offsets){
	int tid=get_global_id(0);
	int j=get_global_id(1);
	//float2 half=split(0.5f);
	float2 jj=split(__int2float_rn(j));
	float2 tmp=df64_mult(periods[tid],jj);
	//for(unsigned j=0;j<n_per[tid];j++)
	float int_part=trunc(tmp.x);
	tmp.x-=int_part;
	tmp.x+=tmp.y;
	offsets[tid+j*get_global_size(0)]=((int)int_part)+((int)rint(tmp.x));
}
#endif